

##########################################################################
#
# This module provides a class for interacting with this the Sloan-Kettering
# Cancer Bioinfomatics Portal Web serice. Can also be run from the command line
#
###########################################################################
'''
CBioPortal.py - Interface with the Sloan-Kettering cBioPortal webservice
========================================================================

:Author: Ian Sudbery
:Release:  $Id$
:Date: |today|
:Tags: Python

The Sloan Kettering cBioPortal webservice provides access to a database of
results of genomics experiments on various cancers. The database is organised
into studies, each study contains a number of case lists, where each list contains
the ids of a set of patients, and genetic profiles, each of which represents an assay
conducted on the patients in the case list as part of the study.

The main class here is the CBioPortal class representing a connection to the cBioPortal
Database. Query's are represented as methods of the class. Study ids or names or case lists
can be provided to the constructor to the object, via the setDefaultStudy and 
setDefaultCaseList methods or to the indevidual query methods. Where ever possible
the validity of parameters is checked *before* the query is executed.

Whenever a query requires a genetic  profile id or a list of such ids, but none are
given, the list of all profiles for which the show_in_analysis flag is set will be
used. 

All of the commands provided in the webservice are implemented here and as far as 
possible the name, syntax and paramter names of the query are identical to the
raw commands to the webservice. These queries are:

* getCancerStudies,
* getCaseLists,
* getProfileData,
* getMutationData,
* getClinicalData,
* getProteinArrayInfo,
* getProteinArrayData,
* getLink,
* getOncoprintHTML.

In addition two new queries are implememented that are not part of the webservice:

* getPercentAltered and
* getTotalAltered

These emulate the function of the website where the percent of cases that show any
alteration for the gene and profiles given are returned (getPercentAltered, or the
percent of cases that show an alteration in any of the genes (getTotalAltered) is
returned. 

examples::

   gene_list = [ "TP53",
   "BCL2",
   "MYC"  ]
   portal = CBioPortal()
   portal.setDefaultStudy(study = "prad_mskcc")
   portal.setDefaultCaseList(case_set_id = "prad_all_complete")
   portal.getPercentAltered(gene_list = gene_list)

or more tersely::

   portal.CBioProtal()
   portal.getPercentAltered(study = "prad_mskcc", case_set_id = "prad_all_complete", 
                            gene_list = ["TP53","BCL2","MYC"], 
                            genetic_profile_id =["prad_mskcc_mrna"])

Any warnings returned by the query are stored in CBioPortal.last_warnings.

Query's that would give too long an URL are split into smaller querys and the results
combined transparently.

A commandline interface is provided for convenience, syntax::

   python CBioPortal.py [options] command(s)
'''
import urllib2, re, optparse, sys, os, IOTools
from collections import OrderedDict as odict
import Experiment as E
 
#######################################################################################
#######################################################################################
#######################################################################################


class CBioPortal():
    
    url = "http://www.cbioportal.org/public-portal/webservice.do"
    

    def __init__(self, url = None, study = None, study_name = None, case_list_id = None):
        ''' connect to the cBioPortal Database. If no url is specified the default url
        is used. A list of of valid study ids is retrieved from the database. This
        both confirms that the datavase is reachable, and provides cached checking for the
        ids provided. If a study or study name is provided then this is set as the defualt
        study for this session and the details of the availible profiles and cases is retrieved.
        'Study' is the study id. If both study and study_name are specified then the 
        study id is used '''

        if url:
            self.url = url

        self.getCancerStudies()
        self.study = None
        self.case_list = None

        if study:
            
            if study in self._valid_study_ids:
                self.study = study
            
                self.profiles = self.getGeneticProfiles(study)
                self.cases = self.getCaseLists(study)
            else:
                raise ValueError("%s is not a valid study id" % study)
        
        elif study_name:
            if study_name in [x['name'] for x in cancer_studies]:
                study = [x['cancer_study_id'] for x in cancer_studies if x['name'] == study_name][0]
                self.study = study
                self.profiles = self.getGeneticProfiles(study)
                self.cases = self.getCaseLists(study)

            else:
                raise ValueError("%s is not a valid study name" % study_name)
        
        if case_list_id:
            self.setDefaultCaseList(case_list_id)

###############################################################################################

    def _executeQuery(self,command,args = None):
        '''execute the provided command on the database. args are specified as a dictionary.
        error checking on the results is performed here and the returned value is a list of
        lists with each list representing a row of the returned table. '''

        
        try:
            argument_string = "&".join(["%s=%s" % (x,args[x]) for x in args])
            command_string = "&".join([command,argument_string])
        except TypeError:
            command_string = command

        
        query = "%s?cmd=%s" % (self.url, command_string)

        if len(query) > 2048:
            if "gene_list" in args:
                genes = args['gene_list'].split(",")
                
                if len(genes) < 2:
                    raise ValueError("Request too long")

               
                args['gene_list'] = ",".join(genes[(len(genes)/2):])
                query1 = self._executeQuery(command,args)
                warnings = self.last_warnings
                
                args['gene_list'] = ",".join(genes[:(len(genes)/2)])
                query2 = self._executeQuery(command,args)
                self.last_warnings = warnings + self.last_warnings

                return query1 + query2

                            
        data = urllib2.urlopen(query)

        
        line = data.readline()
        self.last_query = query
        self.last_status = line
        self.last_warnings = []
        self.last_header = [self.last_status]
        return_table = []

        while re.match("^#",line):

            if re.match("^# Warning: (.+)",line):
                self.last_warnings.append(re.match("^# Warning: (.+)",line).groups(1)[0])
                self.last_header.append(line)
                line = data.readline()
                continue

            elif re.match("^#", line):
                self.last_header.append(line)
                line = data.readline()
                continue

        if re.match("^Error: (.+)", line):
                self.last_header.append(line)
                raise CDGSError(re.match("^Error: (.+)",line).groups(1)[0],query)
        line = line.strip()
        headers = line.split("\t")

        for line in data:
            if re.match("^# Warning: (.+)",line):
                self.last_warnings.append(re.match("^# Warning: (.+)",line).groups(1)[0])
                self.last_header.append(line)
                line = data.readline()
                continue
            line = line.strip()
            return_table.append(odict(zip(headers,line.split("\t"))))
      
        return return_table
###########################################################################################
                                   
    def _getStudyId(self,study,study_name):
        
        if study:
            if study in self._valid_study_ids:
                return study
            else:
                raise ValueError("%s is not a valid study id" % study)

        elif study_name:
            name_lookup = [x['study'] for x in self.cancer_studies if x['name'] == study_name]
            if len(name_lookup) == 1:
                return name_lookup[0]
            else:
                raise ValueError("Cannot find study id for study '%s'" % study_name)
        
        elif self.study:
                return self.study
        
        else:
            return None

###########################################################################################

    def getCancerStudies(self):
        ''' Fetchs the list of cancer studies currently in the database. Returns list
        of dictionaries with three entries 'cancer_study_id','name' and 'description'.
        Also caches this data to verify the validity of later calls '''
        
        cancer_studies = self._executeQuery("getCancerStudies")
        self.cancer_studies = cancer_studies
        self._valid_study_ids = [x['cancer_study_id'] for x in cancer_studies]

        return self.cancer_studies

###########################################################################################

    def getGeneticProfiles(self,study = None, study_name = None):
        ''' Fetches the valid genetic profiles for a particular study. study is the study id.
        If both study and study_name are specified, study is used. If neither study nor study
        name is specified then the default study is used if set, if not a value error is raised.
        Returns a list of dictionaries '''
        
        study= self._getStudyId(study,study_name)

        if not study:
            raise ValueError("A study must be specified if no default is set")

        genetic_studies = self._executeQuery(command = "getGeneticProfiles",
                                             args = dict(cancer_study_id=study) )
        return genetic_studies
############################################################################################

    def getCaseLists(self, study = None, study_name = None):
        ''' Retrieves meta-data regarding all case lists stored about a specific cancer study.
        For example, a within a particular study, only some cases may have sequence data, and
        another subset of cases may have been sequenced and treated with a specific therapeutic
        protocol. Multiple case lists may be associated with each cancer study, and this method
        enables you to retrieve meta-data regarding all of these case lists. 

        Data is returned as a list of dictionaries with the following entries:
         * case_list_id: a unique ID used to identify the case list ID in subsequent
           interface calls. This is a human readable ID. For example, "gbm_all" identifies
           all cases profiles in the TCGA GBM study.
           
         * case_list_name: short name for the case list.

         * case_list_description: short description of the case list.

         * cancer_study_id: cancer study ID tied to this genetic profile. Will match the
           input cancer_study_id.
         * case_ids: space delimited list of all case IDs that make up this case list. '''

        study = self._getStudyId(study,study_name)
        
        if not study:
            raise ValueError("A study must be specified if no default is set")

        case_lists = self._executeQuery(command = "getCaseLists",
                                   args=dict({'cancer_study_id': study}))
        return case_lists

###############################################################################################
        
    def getProfileData(self, gene_list, case_set_id = None, genetic_profile_id = None, study = None, study_name = None):
        ''' Retrieves genomic profile data for one or more genes
        You can specify one gene and many profiles or one profile and many genes. 
        If you specify no genetic profiles then all genetic profiles for the specified or 
        default study are used if the case_set_id is from that study otherwise a ValueError 
        is raised.
 
        Return value depends on the parameters. If you specify a single genetic profile
        and multiple genes a list of ordered dictionaries with the following entires:
        
            gene_id: Entrez Gene ID
            common: HUGO Gene Symbol
            entries 3 - N: Data for each case

        If you specify multi genetic prfiles and a single gene, a list of ordered dictoraries
        with the following entries is returned:

            genetic_profile_id: The Genetic Profile ID.
            alteration_type: The Genetic Alteration Type, e.g. MUTATION, MUTATION_EXTENDED, COPY_NUMBER_ALTERATION, or MRNA_EXPRESSION.
            gene_id: Entrez Gene ID.
            common: HUGO Gene Symbol.
            Columns 5 - N: Data for each case. '''

        #Do some pre-query checking.
        
        if len(gene_list) > 1 and len(genetic_profile_id) > 1:
            raise ValueError("%i genes and %i profiles specified\n.Please specify either one gene many profiles or one profile many genes" % (len(gene_list),len(genetic_profile_id)))

        study_id = self._getStudyId(study,study_name)
        case_set_id = self._getCaseListId(case_set_id,study_id)

        genetic_profile_id = self._getAndCheckGeneticProfiles(genetic_profile_id = genetic_profile_id, study = study)

    
        gene_list = ",".join(gene_list)
        genetic_profile_id = ",".join(genetic_profile_id)

        profile_data = self._executeQuery( command = "getProfileData",
                                     args = { 'case_set_id': case_set_id,
                                              'genetic_profile_id': genetic_profile_id,
                                              'gene_list': gene_list } )

        return profile_data

###################################################################################################

    def getMutationData(self, gene_list, genetic_profile_id, case_set_id = None, study = None, study_name = None):
        '''For data of type EXTENDED_MUTATION, you can request the full set of annotated
        extended mutation data. This enables you to, for example, determine which sequencing 
        center sequenced the mutation, the amino acid change that results from the mutation,
        or gather links to predicted functional consequences of the mutation.

        Query Format

        
            case_set_id= [case set ID] (required)
            genetic_profile_id= [a single genetic profile IDs] (required).
            gene_list= [one or more genes, specified as HUGO Gene Symbols or Entrez Gene IDs](required)
            

        Response Format

        A list of dictionaries with the following entires
        
            entrez_gene_id: Entrez Gene ID.
            gene_symbol: HUGO Gene Symbol.
            case_id: Case ID.
            sequencing_center: Sequencer Center responsible for identifying this mutation. 
                               For example: broad.mit.edu.
            mutation_status: somatic or germline mutation status. all mutations returned will be 
                               of type somatic.
            mutation_type: mutation type, such as nonsense, missense, or frameshift_ins.
            validation_status: validation status. Usually valid, invalid, or unknown.
            amino_acid_change: amino acid change resulting from the mutation.
            functional_impact_score: predicted functional impact score, as predicted by: Mutation 
                               Assessor.
            xvar_link: Link to the Mutation Assessor web site.
            xvar_link_pdb: Link to the Protein Data Bank (PDB) View within Mutation Assessor web site.
            xvar_link_msa: Link the Multiple Sequence Alignment (MSA) view within the
                            Mutation Assessor web site.
            chr: chromosome where mutation occurs.
            start_position: start position of mutation.
            end_position: end position of mutation.

       If a default study is set then a check will be performed to set if the supplied case id is from
       the specified study. The study can be over written using the study and study_name parameters '''

            
        study_id = self._getStudyId(study,study_name)

        case_set_id = self._getCaseListId(case_set_id,study_id)

        if study_id:
            if not study_id == self.study:
                profiles = getGeneticProfiles(study_id)
            else:
                profiles = self.profiles
                
            if not genetic_profile_id in proiles:
                raise ValueError("%s not a valid genetic profile for study %s" % (genetic_profile_id,gene_id) )
        genetic_profile_id = ",".join(genetic_profile_id)
        gene_list = ",".join(gene_list)
        mutation_data = self._executeQuery(command = "getMutationData",
                                      args = dict({"case_set_id": case_set_id,
                                                   "genetic_profile_id": genetic_profile_id,
                                                   "gene_list": gene_list } ) )
        return mutation_data

###################################################################################################

    def getClinicalData(self, case_set_id = None, study = None, study_name = None):
        ''' Retrieves overall survival, disease free survival and age at diagnosis for specified
        cases. Due to patient privacy restrictions, no other clinical data is available.
        
        Query Format
        ------------
        
            case_set_id= [case set ID] (required)

        Response Format
        ---------------

        A list of dictionaries with the following entries:

            case_id: Unique Case Identifier.
            overall_survival_months: Overall survival, in months.
            overall_survival_status: Overall survival status, usually indicated as "LIVING" or "DECEASED".
            disease_free_survival_months: Disease free survival, in months.
            disease_free_survival_status: Disease free survival status, usually indicated as "DiseaseFree" or "Recurred/Progressed".
            age_at_diagnosis: Age at diagnosis.

        If a study is specified or a defualt study is set, then the case_set_id will be tested to 
        check if it exists for that study. '''

        
        study_id = self._getStudyId(study,study_name)

        case_set_id = self._getCaseListId(case_set_id, study_id)

        clincal_data = self._executeQuery(command = "getClinicalData",
                                     args = dict({'case_set_id': case_set_id}))
        return clincal_data

###################################################################################################

    def getProteinArrayInfo(self,protein_array_type = None, gene_list=None, study = None, study_name = None):
        ''' Retrieves information on antibodies used by reverse-phase protein arrays (RPPA) to
        measure protein/phosphoprotein levels.

        Query Format
        ------------
        
            cancer_study_id= [cancer study ID] (required)
            protein_array_type= [protein_level or phosphorylation]
            gene_list= [one or more genes, specified as HUGO Gene Symbols or Entrez Gene IDs]. 

        Response Format
        ---------------

        A list of dictionaries with the following entires:

            ARRAY_ID: The protein array ID.
            ARRAY_TYPE: The protein array antibody type, i.e. protein_level or phosphorylation.
            GENE: The targeted gene name (HUGO gene symbol).
            RESIDUE: The targeted resdue(s).

        If no study is specified the default study is used. If that is not specified an error is 
        raised. '''

        study = self._getStudyId(study,study_name)
        args = dict({'cancer_study_id': study})

        if gene_list:
            args['gene_list'] =  ",".join(gene_list)

        if protein_array_type:
            args['protein_array_type'] = protein_array_type

        
        protein_array_info = self._executeQuery(command = "getProteinArrayInfo",args = args)
                                                  
        
        return protein_array_info

###################################################################################################

    def getProteinArrayData(self, protein_array_id = None, case_set_id = None, array_info = 0, study = None, study_name = None):
        ''' Retrieves protein and/or phosphoprotein levels measured by reverse-phase protein
        arrays (RPPA).
        
        Query Format
        ------------

        case_set_id= [case set ID]
        protein_array_id= [one or more protein array IDs] as list. 
        array_info= [1 or 0]. If 1, antibody information will also be exported.

        Response Format 1
        -----------------

        If the parameter of array_info is not specified or it is not 1, returns a list of dictionaries
        with the following columns.

        ARRAY_ID: The protein array ID.
        Columns 2 - N: Data for each case.

        Response Format 2
        -----------------

        If the parameter of array_info is 1, you will receive a list of ordered dictionaries with the 
        following entires:

        ARRAY_ID: The protein array ID.
        ARRAY_TYPE: The protein array antibody type, i.e. protein_level or phosphorylation.
        GENE: The targeted gene name (HUGO gene symbol).
        RESIDUE: The targeted resdue(s).
        Columns 5 - N: Data for each case.

        If the defualt study is set then the case_set_id will be check. The default study can be
        overidden using the study or study_name parameters. '''

        
        study_id = self._getStudyId(study,study_name)

        case_set_id = self._getCaseListId(case_set_id,study_id)
        args = dict({"case_set_id": case_set_id,
                     "array_info": array_info })

        if protein_array_id:
            args['protein_array_id'] = ",".join(protein_array_id)

        protein_array_data = self._executeQuery(command = "getProteinArrayData",args = args)
                                           
        return protein_array_data

###################################################################################################

    def getLink(self,gene_list, study = None, study_name = None,  report = "full"):
        '''return a perminant link to the cBioPortal report for the gene_list
            cancer_study_id=[cancer study ID]
            gene_list=[a comma separated list of HUGO gene symbols] (required)
            report=[report to display; can be one of: full (default), oncoprint_html] '''

        study = self._getStudyId(study,study_name)
        if not study:
            raise ValueError("Study must be specified")
        if not report in ["full", "oncoprint_html"]:
            raise ValueError("%s is not a valid report" % report)
        
        url = "/".join(self.url.split("/")[:-1])
        
        gene_list = ",".join(gene_list)
        return "%s/link.do?cancer_study_id=%s&gene_list=%s&report=%s" % (url,study,gene_list,report)

###################################################################################################

    def getOncoprintHTML(self,gene_list, study = None, study_name = None):
        ''' returns the HTML for the oncoprint report for the specified gene list and study '''

        url = "/".join(self.url.split("/")[0:-1])
        gene_list = ",".join(gene_list)
        command = "%s/link.do?cancer_study_id=%s&gene_list=%s&report=oncoprint_html" % (url,study,gene_list)
        return urllib2.urlopen(command).read()

###################################################################################################

    def setDefaultStudy(self, study = None, study_name = None):
        ''' sets a new study as the default study. Will check that the study id is valid '''
        study = self._getStudyId(study,study_name)
        self.study = study
        self.profiles = self.getGeneticProfiles()
        self.cases = self.getCaseLists()

        all_case_list = [x['case_list_id'] for x in self.cases if x['case_list_name'] == "All Tumours"]

        if len (all_case_list) == 1:
            self.setDefaultCaseList(all_case_list)

##################################################################################################

    def setDefaultCaseList(self, case_set_id, study = None, study_name = None):
        ''' set the default case list. If study is not specified the default study will be used.
        The study will be used to check that the case_set exists '''

        study = self._getStudyId(study, study_name)
        case_list = self._getCaseListId(case_set_id,study = study)
        
        if not study == self.study:
            self.setDefaultStudy(study)

        self.case_list = case_list

###################################################################################################

    def _getCaseListId(self,case_set_id = None, study = None, strict = True):
        ''' checking is only done if study is specified or a default is set '''

        study_id = self._getStudyId(study,None)
        
        if case_set_id:
            if study_id:
                if not study_id == self.study:
                    case_lists = [x['case_list_id'] for x in self.getCaseLists(study_id)]
                else:
                    case_lists = [x['case_list_id'] for x in self.cases]
                    
                if not case_set_id in case_lists:
                    raise ValueError("%s is not a valid case list for study %s" % (case_set_id,study_id))
            return case_set_id
           
                
        else:
            if self.case_list:
                return self.case_list
            else:
                raise ValueError("No case_set_id provided and no default set")
 
##############################################################################################
  
    def _getAndCheckGeneticProfiles(self,genetic_profile_id = None, study = None):

        study_id = self._getStudyId(study,None)
        if not genetic_profile_id:
            if not study_id:
                raise ValueError("Either genetic_profile_id or study must be specified")

            if study_id == self.study:
                genetic_profile_id = [x['genetic_profile_id'] 
                                      for x in self.profiles 
                                      if x['show_profile_in_analysis_tab'] == "true"]
            else:
                genetic_profile_id = [x['genetic_profile_id'] 
                                      for x in self.getGeneticProfiles(study_id)
                                      if c['show_profile_in_analysis_tab'] == "true"]

            return genetic_profile_id
        
        else:
            
            if not study_id:
                return genetic_profile_id

            if study_id == self.study:
                genetic_profile_id = [x['genetic_profile_id'] for x in self.profiles
                                      if x['genetic_profile_id'] in genetic_profile_id ]
            else:
                genetic_profile_id = [x['genetic_profile_id'] for x in self.getGeneticProfiles(study_id)
                                      if x['genetic_profile_id'] in genetic_profile_id]

            if len(genetic_profile_id) == 0:
                raise ValueError("no valid genetic_profile_ids found")
            
            return genetic_profile_id
###################################################################################################    

    def getPercentAltered(self,gene_list, study = None, study_name = None, case_set_id = None, genetic_profile_id = None, threshold = 2):
        '''Get the percent of cases that have one or more of the specified alterations for each gene

        Query Format
        ------------

        study = [cancer_study_id] The study to use.
        study_name = [cancer_study_name] The name of the study to use. If neither this nor
                     study are specified, then the default is used.
        case_set_id = [case_set_id] The case list to use. If not specified, the default case list is
                      used. 
        gene_list = [one or more genes, specified as HUGO Gene Symobls or ENtrez Gene IDs] (require)
        genetic_profile_id = [one or more genetic profile IDs] If none specified all genetic profiles
        for the specified study are used..
        threhold = [z_score_threshold] the numeric threshold at which a mrna expression z-score is 
        said to be significant.

        Response Format
        ---------------

        A list of dictionaries with the following entries
        gene_id: The Entrez Gene ID
        common: The Hugo Gene Symbol
        altered_in: The percent of cases in which the gene is altered 

        One implementation note is that a guess must be made as to wether a returned profile
        value represents a alteration or not. Currently guesses are only made for copy number
        variation, mrna expression and mutionation'''

        study = self._getStudyId(study,study_name)
        case_set_id = self._getCaseListId(case_set_id, study)
        genetic_profile_id = self._getAndCheckGeneticProfiles(genetic_profile_id, study)

        if study and study == self.study:
            profiles = self.profiles
            case_list = [x['case_ids'] for x in self.cases if x['case_list_id'] == case_set_id][0]

        else:
            profiles = self.getGeneticProfiles(study)
            case_list = [x['case_ids'] for x in  self.getCaseLists(study) if x['case_list_id'] == case_set_id][0]

        return_table = []
        case_list = case_list.split(" ")
 
        data = []
        warnings = []
        for profile in genetic_profile_id:
            
            data.append(self.getProfileData(gene_list = gene_list, case_set_id = case_set_id, genetic_profile_id = [profile]))
            warnings.extend(self.last_warnings)
        
        #data[profile][gene][case]
        for gene in range(len(data[0])):
            cases_altered = 0.0
            
            geneProfile = [x[gene] for x in data]
            
            for case in case_list:
                
                if len([geneProfile[x][case] for x in range(len(geneProfile)) 
                        if self._guessAlteration(geneProfile[x][case],genetic_profile_id[x],profiles)]) > 0:
                              
                    cases_altered += 1
               
            
            return_table.append(dict({'gene_id': geneProfile[0]['GENE_ID'],
                                          'common': geneProfile[0]['COMMON'],
                                          'altered_in': cases_altered*100/len(case_list) }) )

        self.last_warnings = warnings
        return return_table

#############################################################################################

    def getTotalAltered(self,gene_list, study = None, study_name = None, case_set_id = None, genetic_profile_id = None, threshold = 2):
        ''' Calculate the percent of cases in which any one of the specified genes are altered '''


        study = self._getStudyId(study,study_name)
        case_set_id = self._getCaseListId(case_set_id, study)
        genetic_profile_id = self._getAndCheckGeneticProfiles(genetic_profile_id, study)

        if study == self.study:
            profiles = self.profiles
            case_list = [x['case_ids'] for x in self.cases if x['case_list_id'] == case_set_id][0]
        else:
            profiles = self.getGeneticProfiles(study)
            case_list = [x['case_ids'] for x in  self.getCaseLists(study) if x['case_list_id'] == case_set_id][0]
        case_list = case_list.split(" ")

        data = []


        #data[gene][profile][case]
        cases_altered = 0.0
        for gene in gene_list:
            data.append(self.getProfileData(gene_list = [gene], case_set_id = case_set_id, genetic_profile_id = genetic_profile_id))
            #catch special case where only a single genetic_profile_id is passed in, so the query is single
            #gene and single profile and returns genes as columns format.
            if len(genetic_profile_id) == 1 and len(data[0]) == 1:
                data[0][0]['GENETIC_PROFILE_ID'] = genetic_profile_id[0]
            
        for case_id in case_list:
            
            case_altered = False
            for gene in data:

                
                altered = len([x for x in gene 
                              if self._guessAlteration(x[case_id],x['GENETIC_PROFILE_ID'], profiles, threshold)])
                if (altered > 0):
                    case_altered =True
            
            if case_altered == True:
                cases_altered += 1.0

        return cases_altered*100/len(case_list)

#################################################################################################
       
            
        
      
    def _guessAlteration(self,value,genetic_profile_id, genetic_profiles, threshold = 2):
        
        alteration_type = [x['genetic_alteration_type'] 
                           for x in genetic_profiles if x['genetic_profile_id'] ==  genetic_profile_id][0]

        #print alteration_type                  
        if alteration_type == "COPY_NUMBER_ALTERATION":
            if value == "0" or value == "-1" or value == "1":
                return False
            elif value == "NaN":
                return False
            else:
                return True

        elif alteration_type == "MRNA_EXPRESSION":
            if re.search("[^0-9\.\-]", value):
                #is character string
                return False
            elif re.search("[0-9]+\.[0-9]+",value):
                #is float
                value = float(value)
                if abs(value) > threshold:
                    return True
                else:
                    return False
            else: # must be int?
                #print value
                if value == "0":
                    return False
                else:
                    return True

        elif alteration_type == "METHYLATION":
            return False

        elif alteration_type == "MUTATION" or alteration_type == "MUTATION_EXTENDED":
            #print "hello"
            if value == "NaN":
                return False
            else:
                return True
        else:
            return False

        

                            
        

class CDGSError(Exception):
    '''exception that handles errors returned by querys in the database'''
    def __init__(self, error, request):
        self.error = error
        self.request = request
    def __str__(self):
        return "Request %s return error:\n%s" % (self.request,self.error)


def tableToString(intable):

    headers = "\t".join([x for x in intable[0]])
    
    line_list = []
    for line in intable:

        line_list.append("\t".join([ str(line[x]) for x in line]))

    outTable = "\n".join(line_list)
    outTable = headers + "\n" + outTable

    return outTable


def main(argv = None):

    parser = E.OptionParser(version = "%prog version: $Id: CBioPortal.py 2888 2012-06-07 15:52:00Z ians $", usage = globals()["__doc__"])

    parser.add_option("-o","--output_file", type="string", default = None,
                      help="[Optional] Filename to output results to. [default=STDOUT]" )
    parser.add_option("-u","--url",type="string",default="http://www.cbioportal.org/public-portal/webservice.do",
                      help="[Optional] Url to the cBioPortal webservice [default=%default]" )

    cqueryopts = optparse.OptionGroup(parser,"Common parameters","Common arguments to the query")
    cqueryopts.add_option("-s", "--study_id", dest="study_id", type="string", default = None,
                      help="[Required/OPtional]  cBioPortal ID for study [default=%default].\n This or study_name required for: getGeneticProfiles, getCaseLists, getProteinArrayInfo, getLink,getOncoprintHTML, getPercentAltered, getTotalAltered"  )
    cqueryopts.add_option("-n", "--study_name",dest = "study_name", type = "string", default = None,
                      help="[Required/Optional] cBioPortal Name for study [defualt=%default].\n See above for which commands require this.")
    cqueryopts.add_option("-c", "--case_set_id", dest="case_set_id", type="string", default = None,
                      help="[Required for some] cBioPortal case_set_id specifying the case list to use.\nRequired for getProfileData, getMutationData, getClincalData, getProteinArrayData, getPercentAltered, getTotalAltered. Default is case_set_id for case list 'All Tumours' ")
    cqueryopts.add_option("-g", "--gene_list", dest = "gene_list", type = "string", default = None,
                      help="[Required for some] Comma seperated list of HUGO gene symbols or Entrez gene IDs.\nRequired for getProfileData, getMutationData, getLink, getOncoprintHTML" )
    cqueryopts.add_option("-f","--gene_list_file", dest = "gene_list_file", type="string", default = None,
                          help="[Optional] Filename to read in gene_list from" )
    cqueryopts.add_option("-p", "--profile_id", dest = "profile_id", type = "string",
                      help="[Optional] Comma seperated list of cBioPortal genetic_profile_ids. If none are specified then the list of profiles for the study where display in analysis is True is used." )
    


    squeryopts = optparse.OptionGroup(parser,"Query specific parameters", "Arguments specific to a particular query")
    squeryopts.add_option("--protein_array_type", dest="protein_array_type", type="string", default = "protein_level",
                      help="[Optional] Either protein_level or phosphorylation [default=%default]" )
    squeryopts.add_option("--protein_array_id", dest = "protein_array_id", type = "string",
                      help="[Required for some] comma seperated list of one or more protein array IDs" )
    squeryopts.add_option("--array_info", dest ="protein_array_info", type ="int",  default = 0,  
                      help="[Optional] If 1, antibody infomation will also be exported in a getProteinArrayData query [default=%default]" )
    squeryopts.add_option("--report", dest = "report", type = "string", default = "full",
                      help = "[Optional] Report type to display for getLink. Either full or oncoprint_html [default=%default] " )
    squeryopts.add_option("--threshold", dest = "threshold", type="int", default = 2, 
                      help = "[Optional] Threshold for deciding if an alteration is significant for continuous metrics [default=%default]" )
    
    parser.add_option_group(cqueryopts)
    parser.add_option_group(squeryopts)

    (options,args) = E.Start(parser, add_pipe_options = False, add_output_options = False, argv = argv)

    portal = CBioPortal(url = options.url, study = options.study_id, study_name = options.study_name, case_list_id = options.case_set_id)

    results = []

    if options.gene_list_file:
        infile = IOTools.openFile(options.gene_list_file)
        gene_list = [x.strip() for x in infile]
    elif options.gene_list:
        gene_list = options.gene_list.split(",")

    if options.profile_id:
        profile_id = options.profile_id.split(",")
    else:
        profile_id = None

    if "getCancerStudies" in args:
        results.append(portal.getCancerStudies())
    
    if "getGeneticProfiles" in args:
        results.append(portal.getGeneticProfiles())

    if "getCaseLists" in args:
        results.append(portal.getCaseLists())

    if "getProfileData" in args:
        results.append(portal.getProfileData(gene_list = gene_list, genetic_profile_id = profile_id))

    if "getMutationData" in args:
        results.append(portal.getMutationData(gene_list = gene_list, genetic_profile_id = profile_id))

    if "getClinicalData" in args:
        results.append(portal.getClinicalData())

    if "getProteinArrayInfo" in args:
        results.append(portal.getProteinArrayInfo(gene_list = gene_list, protein_array_type = options.protein_array_type))

    if "getProteinArrayData" in args:
        results.append(portal.getProteinArrayData(protein_array_id = options.protein_array_id, array_info =  options.array_info))

    if "getPercentAltered" in args:
        results.append(portal.getPercentAltered(gene_list = gene_list, genetic_profile_id = profile_id, threshold = options.threshold))

    if "getLink" in args:
        results.append(portal.getLink(gene_list = gene_list, report = options.report))

    if "getOncoprintHTML" in args:
        results.append(portal.getOncoprintHTML(gene_list = gene_list))
    
    if len(results) == 0:
        sys.stderr.write( "No recognised query commands provided")
        sys.exit()
    
    if options.output_file:
        outf = IOTools.openFile(options.output_file, "w")
    else:
        outf = sys.stdout

    
    for result in results:
        try:
            outf.write(tableToString(result))
        except:
            outf.write(result)


    E.Stop()

    

           
if __name__ == "__main__":
    sys.exit(main(sys.argv))



################################################################################
#   Gene prediction pipeline 
#
#   $Id: matrix2stats.py 2795 2009-09-16 15:29:23Z andreas $
#
#   Copyright (C) 2004 Andreas Heger
#
#   This program is free software; you can redistribute it and/or
#   modify it under the terms of the GNU General Public License
#   as published by the Free Software Foundation; either version 2
#   of the License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#################################################################################
'''
matrix2stats.py - compute statistics on matrices
================================================

:Author: Andreas Heger
:Release: $Id: matrix2stats.py 2795 2009-09-16 15:29:23Z andreas $
:Date: |today|
:Tags: Python

Purpose
-------

compute statistics on matrices.

Chi-Squared test
++++++++++++++++

Compute the Chi-squared test on a contingency table.
If --pairwise is given, the pairwise

The command line option *iteration* denotes what values will
be computed:

   pairwise
      compute all pairs assuming symmetry: `((1,2), (1,3), (2,3))`

   all-vs-all
      compute all-vs-all pairwise explicetely: `((1,2), (1,3), (2,1), (2,3), (3,1), (3,2))`

Usage
-----

Type::

   python <script_name>.py --help

for command line help.

Code
----
'''


import os
import sys
import string
import re
import optparse
import math
import StringIO



import numpy
import CGAT.Experiment as E
import CGAT.IOTools as IOTools
import CGAT.MatlabTools as MatlabTools
import CGAT.Stats as Stats
import scipy

if __name__ == "__main__":

    parser = E.OptionParser( version = "%prog version: $Id: matrix2stats.py 2795 2009-09-16 15:29:23Z andreas $",
                                    usage = globals()["__doc__"] )

    parser.add_option("-m", "--method", dest="method", type="choice",
                      choices=("chi-squared", "pearson-chi-squared" ),
                      help="statistical methods to apply." )
    
    parser.add_option("-t", "--headers", dest="headers", action="store_true",
                      help="matrix has row/column headers."  )

    parser.add_option("--no-headers", dest="headers", action="store_false",
                      help="matrix has no row/column headers."  )

    parser.add_option("-i", "--input-format", dest="input_format", type="choice",
                      choices=("full", "sparse", "phylip" ),
                      help="""input format for matrix."""  )

    parser.add_option("-o", "--output-format", dest="output_format", type="choice",
                      choices=("full", "sparse", "phylip" ),
                      help="""output format for matrix."""  )
    
    parser.add_option("-p", "--parameters", dest="parameters", action="append", type="string",
                      help="parameters for various functions." )


    parser.add_option("-a", "--iteration", dest="iteration", type="choice",
                      choices=("pairwise", "all-vs-all" ),
                      help="""how to compute stats [%default]."""  )

    parser.set_defaults(
        method = "chi-squared",
        headers = True,
        value_format = "%6.4f",
        pvalue_format = "%6.4e",
        input_format = "full",        
        write_separators = True,
        parameters = [],
        iteration = None,
        )

    (options, args) = E.Start( parser )
    
    lines = filter( lambda x: x[0] != "#", sys.stdin.readlines())

    chunks = filter( lambda x: lines[x][0] == ">", range(len(lines)) )

    if not chunks:
        options.write_separators = False
        chunks = [-1]
        
    chunks.append( len(lines) )

    ninput, noutput, nskipped = 0, 0, 0

    if options.write_separators:
        options.stdout.write( "test\t" )

    header_prefix = ""

    if options.method == "chi-squared":
        header_prefix = "observed\texpected"
        options.stdout.write( "\t".join( ( header_prefix, "n", "min", "max", "chi", "df", "P", "passed", "phi" ) ) + "\n" )

    elif options.method in ("pearson-chi-squared",):
        options.stdout.write("column\t")
        options.stdout.write( "\t".join( ( header_prefix, "n", "prob", "obs", "exp", "chi", "df", "P", "passed", "phi" ) ) + "\n" )

        if len(options.parameters) == 0:
            raise "out of parameters - please supply probability or filename with probabilities."

        param = options.parameters[0]
        del options.parameters[0]

        if options.write_separators:
            probabilities = IOTools.ReadMap( open( param, "r" ), map_functions = (str,float) )
        else:
            probability = float( param )


    for x in range(len(chunks) -1 ):
        ninput += 1
        matrix, row_headers, col_headers = MatlabTools.readMatrix( StringIO.StringIO("".join(lines[chunks[x]+1:chunks[x+1]])),
                                                                       format=options.input_format, 
                                                                       headers = options.headers )
        nrows, ncols = matrix.shape

        if options.loglevel >= 2:
            options.stdlog.write( "# read matrix: %i x %i, %i row titles, %i colum titles.\n" %\
                                      (nrows, ncols, len(row_headers), len(col_headers)))

        if options.write_separators:
            options.stdout.write( lines[chunks[x]][1:-1] + "\t" )

        pairs = []
        if options.iteration == "pairwise":
            pairs = []
            for row1 in range( 0, len(row_headers) ):
                for row2 in range( row1+1, len(row_headers) ):
                    pairs.append( (row1, row2) )
        elif options.iteration == "all-vs-all":
            pairs = []
            for row1 in range( 0, len(row_headers) ):
                for row2 in range( 0, len(row_headers) ):
                    if row1 == row2: continue
                    pairs.append( (row1, row2) )
    
        if options.method == "chi-squared":
            
            for row1, row2 in pairs:
                row_header1 = row_headers[row1]
                row_header2 = row_headers[row2]
                try:
                    result = Stats.doChiSquaredTest( numpy.vstack( (matrix[row1], matrix[row2] ) ) )
                except ValueError:
                    nskipped += 1
                    continue

                noutput += 1
                options.stdout.write( "\t".join( ( "%s" % row_header1,
                                                   "%s" % row_header2,
                                                   "%i" % result.mSampleSize,
                                                   "%i" % min(matrix.flat),
                                                   "%i" % max(matrix.flat),
                                                   options.value_format % result.mChiSquaredValue,
                                                   "%i" % result.mDegreesFreedom,
                                                   options.pvalue_format % result.mProbability,
                                                   "%s" % result.mSignificance,
                                                   options.value_format % result.mPhi ) ) + "\n" )
            

        elif options.method == "pearson-chi-squared":

            if nrows != 2:
                raise "only implemented for 2xn table"
            
            if options.write_separators:
                id = re.match( "(\S+)", lines[chunks[x]][1:-1] ).groups()[0]
                probability = probabilities[id]
                
            for col in range(ncols):
                options.stdout.write( "%s\t" % col_headers[col] )
                result = Stats.doPearsonChiSquaredTest( probability , sum(matrix[:,col]), matrix[0,col] )
                options.stdout.write( "\t".join( ( "%i" % result.mSampleSize,
                                                   "%f" % probability,
                                                   "%i" % result.mObserved,
                                                   "%f" % result.mExpected,
                                                   options.value_format % result.mChiSquaredValue,
                                                   "%i" % result.mDegreesFreedom,
                                                   options.pvalue_format % result.mProbability,
                                                   "%s" % result.mSignificance,
                                                   options.value_format % result.mPhi ) )  )
                if col < ncols - 1:
                    options.stdout.write("\n")
                    if options.write_separators:
                        options.stdout.write( lines[chunks[x]][1:-1] + "\t" )

            options.stdout.write( "\n" )

    E.info( "# ninput=%i, noutput=%i, nskipped=%i\n" % (ninput, noutput, nskipped) )

    E.Stop()



################################################################################
#
#   MRC FGU Computational Genomics Group
#
#   $Id$
#
#   Copyright (C) 2009 Andreas Heger
#
#   This program is free software; you can redistribute it and/or
#   modify it under the terms of the GNU General Public License
#   as published by the Free Software Foundation; either version 2
#   of the License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#################################################################################
'''
table2table.py - operate on tables
==================================

:Author: Andreas Heger
:Release: $Id$
:Date: |today|
:Tags: Python

Purpose
-------

This script implements a few methods for manipulating tables.

Methods working on all tables:
++++++++++++++++++++++++++++++

transpose
   transpose a table

split-fields
   Split muliple-value fields in each row at ``--separator``. Output
   multiple rows with all combinations.

group
   Group values by column

join-column
   Join rows in a table by columns

expand-table
   If a field in a row contains multiple values,
   the row is expanded into multiple rows such
   that all values have space.

flatten-table
   Output a table as row/column/value tuples.

as-column
   Output table as a single column. Colums in the original table are appended and output.

Methods for numerical columns
+++++++++++++++++++++++++++++

Some methods make only sense for columns containing numerical values.
If a table contains both numerical and non-numerical data, the
numerical columns can be specified by the ``--columns`` option.

normalize-by-value
   divide all cells in a table by a value

multiply-by-value
   multiply all cells in a table by a value

lower-bound
   replace all cells with a value of less than lower bound with the lower
   bound.

upper-bound
   replace all cells with a value of more than upper bound with the upper
   bound.

normalize-by-table
   divide each cell in a table with the corresponding entry in a secondary
   table.

normalize-by-max
   divide table columns by maximum per column

kullback-leibler
   compute kullback-leibler divergence between two columns. Compute
   both D(a||b), D(b||a) and (D(a||b) + D(b||a)) / 2
   
rank
   substitute cells with their ranks in a column

fdr
   compute an FDR over all columns selected. Replace cells with the qvalues.

Usage
-----

Example::

   python table2table.py --help

Type::

   python table2table.py --help

for command line help.

Documentation
-------------

Code
----

'''
import os
import sys
import string
import re
import optparse
import math
import types
import itertools

import CGAT.Experiment as E
import CGAT.CSV as CSV
import CGAT.Stats as Stats
import scipy.stats

def getColumns( fields, columns = "all" ):
    '''return columns to take.'''
    if columns == "all":
        return list(range( len(fields) ))
    elif columns == "all-but-first":
        return list(range( 1, len(fields) ))
    else:
        return map( lambda x: int(x) - 1, columns.split(","))

##########################################################
##########################################################
##########################################################
def readAndTransposeTable( infile, options ):
    """read table from infile and transpose
    """
    rows = []
    if options.transpose_format == "default":

        for line in infile:
            if line[0] == "#" : continue
            rows.append( line[:-1].split("\t") )

    elif options.transpose_format == "separated":
        for line in infile:
            if line[0] == "#": continue
            key, vals = line[:-1].split( "\t" )
            row = [key] + vals.split(options.separator)
            rows.append( row )
        
    ncols = max( map( lambda x: len(x), rows ) )
    nrows = len(rows)

    new_rows = [ [""] * nrows  for x in range(ncols) ]

    for r in range(0, len(rows)):
        for c in range( 0, len(rows[r]) ):
            new_rows[c][r] = rows[r][c]

    if options.set_transpose_field:
        new_rows[0][0] = options.set_transpose_field
 
    for row in new_rows:
        options.stdout.write( "\t".join( row ) + "\n")

##########################################################
##########################################################
##########################################################        
def readAndGroupTable( infile, options ):
    """read table from infile and group.
    """
    
    fields, table  = CSV.ReadTable( infile, with_header = options.has_headers, as_rows = True )

    options.columns = getColumns( fields, options.columns )

    converter = float
    if options.group_function == "min":
        f = min
    elif options.group_function == "max":
        f = max
    elif options.group_function == "sum":
        f = lambda z: reduce( lambda x,y: x+y, z)
    elif options.group_function == "mean":
        f = scipy.mean
    elif options.group_function == "cat":
        f = lambda x: ";".join( [ y for y in x if y != "" ] )
        converter = str
    elif options.group_function == "uniq":
        f = lambda x: ";".join( [ y for y in set(x) if y != "" ] )
        converter = str
    elif options.group_function == "stats":
        f = lambda x: str(Stats.DistributionalParameters(x))
        # update headers
        new_fields = []
        for c in range(len(fields)):
            if c == options.group_column or c not in options.columns:
                new_fields.append( fields[c] )
                continue
            new_fields += list( map(lambda x: "%s_%s" % (fields[c], x), Stats.DistributionalParameters().getHeaders() ) )
        fields = new_fields

    ## convert values to floats (except for group_column)
    ## Delete rows with unconvertable values
    new_table = []
    for row in table:
        skip = False
        for c in options.columns:
            if c == options.group_column: continue
            if row[c] == options.missing_value:
                row[c] = row[c]
            else:
                try:
                    row[c] = converter(row[c])
                except ValueError:
                    skip = True
                    break
        if not skip: new_table.append(row)
    table = new_table

    new_rows = CSV.GroupTable( table,
                               group_column = options.group_column,
                               group_function = f )

    options.stdout.write("\t".join(fields) + "\n")        
    for row in new_rows:
        options.stdout.write( "\t".join( map(str,row) ) + "\n")

##########################################################
##########################################################
##########################################################        
def readAndExpandTable( infile, options ):
    '''splits fields in table at separator. 
    
    If a field in a row contains multiple values,
    the row is expanded into multiple rows such
    that all values have space.
    '''

    fields, table  = CSV.ReadTable( infile, with_header = options.has_headers, as_rows = True )

    options.stdout.write("\t".join(fields) + "\n")
    
    for row in table:

        data = []
        for x in range(len(fields)):
            data.append( row[x].split( options.separator ) )

        nrows = max( [ len(d) for d in data ] )

        for d in data:
            d += [""] * (nrows - len(d))

        for n in range(nrows):
            options.stdout.write( "\t".join( [ d[n] for d in data ] ) + "\n" )

##########################################################
##########################################################
##########################################################        
def computeFDR( infile, options ):
    '''compute FDR on a table.
    '''

    fields, table  = CSV.ReadTable( infile, with_header = options.has_headers, as_rows = True )

    options.stdout.write("\t".join(fields) + "\n")
    
    for row in table:

        data = []
        for x in range(len(fields)):
            data.append( row[x].split( options.separator ) )

        nrows = max( [ len(d) for d in data ] )

        for d in data:
            d += [""] * (nrows - len(d))

        for n in range(nrows):
            options.stdout.write( "\t".join( [ d[n] for d in data ] ) + "\n" )

##########################################################
##########################################################
##########################################################        
def readAndJoinTable( infile, options ):

    fields, table  = CSV.ReadTable( infile, with_header = options.has_headers, as_rows = True )

    join_column = options.join_column - 1
    join_name = options.join_column_name - 1
    
    join_rows = list(set(map( lambda x: x[join_column], table )))
    join_rows.sort()

    join_names = list(set(map( lambda x: x[join_name], table )))
    join_names.sort()

    join_columns = list(set(range(len(fields))).difference( set( (join_column, join_name) ) ))
    join_columns.sort()

    new_table = []
    map_old2new = {}

    map_name2start = {}
    x = 1
    for name in join_names:
        map_name2start[name] = x
        x += len(join_columns)

    row_width = len(join_columns) * len(join_names)
    for x in join_rows:
        map_old2new[x] = len(map_old2new)
        new_row = [ x ,] + [ "na" ] * row_width 
        new_table.append(new_row)

    for row in table:
        row_index = map_old2new[row[join_column]]
        start = map_name2start[row[join_name]]
        for x in join_columns:
            new_table[row_index][start] = row[x]
            start += 1

    ## print new table
    options.stdout.write( fields[join_column] )
    for name in join_names:
        for column in join_columns:
            options.stdout.write( "\t%s%s%s" % (name, options.separator, fields[column]))
    options.stdout.write( "\n" )

    for row in new_table:
        options.stdout.write( "\t".join( row ) + "\n")
            
##########################################################
##########################################################
##########################################################        
if __name__ == "__main__":

    parser = E.OptionParser( version = "%prog version: $Id: table2table.py 2782 2009-09-10 11:40:29Z andreas $")

    parser.add_option("-m", "--method", dest="methods", type="choice", action="append",
                      choices=( "transpose", "normalize-by-max","normalize-by-value","multiply-by-value",
                               "percentile","remove-header","normalize-by-table",
                               "upper-bound","lower-bound","kullback-leibler",
                                "expand","compress", "fdr", "grep" ),
                      help="""actions to perform on table.""")
    
    parser.add_option("-s", "--scale", dest="scale", type="float",
                      help="factor to scale matrix by."  )
    
    parser.add_option("-f", "--format", dest="format", type="string",
                      help="output number format."  )

    parser.add_option("-p", "--parameters", dest="parameters", type="string",
                      help="Parameters for various functions."  )

    parser.add_option("-t", "--headers", dest="has_headers", action="store_true",
                      help="matrix has row/column headers."  )

    parser.add_option("--transpose", dest="transpose", action="store_true",
                      help="transpose table."  )

    parser.add_option("--set-transpose-field", dest="set_transpose_field", type="string",
                      help="set first field (row 1 and col 1) to this value [%default]."  )

    parser.add_option("--transpose-format", dest="transpose_format", type="choice",
                      choices=("default", "separated", ),
                      help="input format of un-transposed table"  )

    parser.add_option("--expand", dest="expand_table", action="store_true",
                      help="expand table - multi-value cells with be expanded over several rows."  )

    parser.add_option("--no-headers", dest="has_headers", action="store_false",
                      help="matrix has no row/column headers."  )

    parser.add_option("--columns", dest="columns", type="string",
                      help="columns to use."  )

    parser.add_option( "--file", dest="file", type="string",
                      help="columns to test from table.",
                      metavar="FILE" )

    parser.add_option("-d", "--delimiter", dest="delimiter", type="string",
                      help="delimiter of columns." ,
                      metavar="DELIM" )

    parser.add_option("-V", "--invert-match", dest="invert_match", action="store_true",
                      help="invert match." )

    parser.add_option("--sort-by-rows", dest="sort_rows", type="string",
                      help="output order for rows."  )

    parser.add_option("-a", "--value", dest="value", type="float",
                      help="value to use for various algorithms."  )

    parser.add_option("--group", dest="group_column", type="int",
                      help="group values by column."  )

    parser.add_option("--group-function", dest="group_function", type="choice",
                      choices=("min", "max", "sum", "mean", "stats", "cat", "uniq"),
                      help="function to group values by."  )

    parser.add_option("--join-table", dest="join_column", type="int",
                      help="join rows in a table by columns."  )

    parser.add_option("--join-column-name", dest="join_column_name", type="int",
                      help="use this column as a prefix."  )

    parser.add_option("--flatten-table", dest="flatten_table", action="store_true",
                      help="flatten table."  )

    parser.add_option("--as-column", dest="as_column", action="store_true",
                      help="output table as a single column."  )

    parser.add_option("--split-fields", dest="split_fields", action="store_true",
                      help="split fields."  )

    parser.add_option("--separator", dest="separator", type="string",
                      help="separator for multi-valued fields [default=%default]."  )

    parser.add_option( "--fdr-method", dest="fdr_method", type="choice",
                      choices = ( "BH", "bonferroni", "holm", "hommel", "hochberg", "BY" ),
                      help="method to perform multiple testing correction by controlling the fdr [default=%default]."  )
    #IMS: add option to use a column as the row id in flatten
    parser.add_option("--id-column", dest="id_column", type ="string",
                      help="list of column(s) to use as the row id when flattening the table. If None, then row number is used. [defualt=%default].")
    parser.add_option("--variable-name", dest="variable_name", type = "string",
                      help="the column header for the 'variable' column when flattening [default=%default].")
    parser.add_option("--value-name", dest="value_name", type = "string",
                      help="the column header for the 'value' column when flattening [default=%default].")


    parser.set_defaults(
        methods = [],
        scale = 1.0,
        has_headers = True,
        format = "%5.2f",
        value = 0.0,
        parameters = "",
        columns = "all",
        transpose = False,
        set_transpose_field = None,
        transpose_format = "default",
        group = False,
        group_column = 0,
        group_function = "mean",
        missing_value = "na",
        sort_rows = None,
        flatten_table= False,
        separator = ";",
        expand = False,
        join_column = None,
        join_column_name = None,
        compute_fdr = None,
        as_column = False,
        fdr_method= "BH",
        id_column=None,
        variable_name="column",
        value_name="value",
        file=None,
        delimiter = "\t",
        invert_match = False,
        )
    
    (options, args) = E.Start( parser, add_pipe_options = True )

    options.parameters = options.parameters.split(",")
    
    if options.group_column:
        options.group = True
        options.group_column -= 1

    ######################################################################
    ######################################################################
    ######################################################################
    ## if only to remove header, do this quickly
    if options.methods== ["remove-header"]:
        
        first = True
        for line in options.stdin:
            if line[0] == "#": continue
            if first:
                first = False
                continue
            options.stdout.write( line )

    elif options.transpose or "transpose" in options.methods:

        readAndTransposeTable( options.stdin, options )


    elif options.flatten_table:
        #IMS: bug fixed to make work. Also added options for keying on a particular
        #     and adding custom column headings

        fields, table  = CSV.ReadTable( options.stdin, with_header = options.has_headers, as_rows = True )
        
        options.columns = getColumns( fields, options.columns )
        
        if options.id_column:
            id_columns = map(lambda x: int(x) -1,options.id_column.split(","))
            id_header = "\t".join([fields[id_column] for id_column in id_columns])
            options.columns = [x for x in options.columns if x not in id_columns]
        else:
            id_header = "row"

        options.stdout.write( "%s\t%s\t%s\n" %(id_header, options.variable_name, options.value_name) )
        
        for x, row in enumerate(table):

            if options.id_column:
                row_id = "\t".join([row[int(x)-1] for x in options.id_column.split(",")])
            else:
                row_id = str(x)

            for y in options.columns:
                options.stdout.write( "%s\t%s\t%s\n" % (row_id,fields[y], row[y] ))

    elif options.as_column:
        
        fields, table  = CSV.ReadTable( options.stdin, with_header = options.has_headers, as_rows = True )
        options.columns = getColumns( fields, options.columns )
        table = zip( *table )
        
        options.stdout.write( "value\n" )
        
        for column in options.columns:
            options.stdout.write("\n".join( table[column] ) + "\n" )

    elif options.split_fields:

        # split comma separated fields
        fields, table  = CSV.ReadTable( options.stdin, 
                                        with_header = options.has_headers, 
                                        as_rows = True )
        

        options.stdout.write( "%s\n" % ("\t".join(fields)))

        for row in table:
            row = [ x.split(options.separator) for x in row ]
            for d in itertools.product( *row ):
                options.stdout.write( "%s\n" % "\t".join( d ) )
            
    elif options.group:
        readAndGroupTable( options.stdin, options )

    elif options.join_column:
        readAndJoinTable( options.stdin, options )

    elif options.expand_table:
        readAndExpandTable( options.stdin, options )

    elif "grep" in options.methods:

        options.columns = map(lambda x: int(x)-1, options.columns.split(","))

        patterns = []

        if options.file:
            infile = open( options.file, "r")
            for line in infile:
                if line[0] == "#": continue
                patterns.append( line[:-1].split(options.delimiter)[0] )
        else:
            patterns=args

        for line in options.stdin:

            data = line[:-1].split(options.delimiter)
            found = False

            for c in options.columns:

                if data[c] in patterns:
                    found = True
                    break

            if (not found and options.invert_match) or (found and not options.invert_match):
                print line[:-1]
    else:

        ######################################################################
        ######################################################################
        ######################################################################
        ## Apply remainder of transformations
        fields, table  = CSV.ReadTable( options.stdin, with_header = options.has_headers, as_rows = False )
        # convert columns to list
        table = [ list(x) for x in table]

        ncols = len(fields)
        nrows = len(table[0])

        E.info( "processing table with %i rows and %i columns" % (nrows, ncols) )

        options.columns = getColumns( fields, options.columns )
        
        ## convert all values to float
        for c in options.columns:
            for r in range(nrows):
                try:
                    table[c][r] = float (table[c][r] )
                except ValueError:
                    continue

        for method in options.methods:

            if method == "normalize-by-value":

                value = float(options.parameters[0])
                del options.parameters[0]

                for c in options.columns:
                    table[c] = map( lambda x: x / value, table[c] )

            elif method == "multiply-by-value":

                value = float(options.parameters[0])
                del options.parameters[0]

                for c in options.columns:
                    table[c] = map( lambda x: x * value, table[c] )

            elif method == "normalize-by-max":

                for c in options.columns:
                    m = max( table[c] )
                    table[c] = map( lambda x: x / m, table[c] )

            elif method == "kullback-leibler":
                options.stdout.write("category1\tcategory2\tkl1\tkl2\tmean\n")
                for x in range(0,len(options.columns)-1):
                    for y in range(x+1, len(options.columns)):
                        c1 = options.columns[x]
                        c2 = options.columns[y]
                        e1 = 0
                        e2 = 0
                        for z in range(nrows):
                            p = table[c1][z]
                            q = table[c2][z]
                            e1 += p * math.log( p / q )
                            e2 += q * math.log( q / p )

                        options.stdout.write("%s\t%s\t%s\t%s\t%s\n" % (fields[c1], fields[c2],
                                                                       options.format % e1,
                                                                       options.format % e2,
                                                                       options.format % ((e1 + e2) / 2)) )
                E.Stop()
                sys.exit(0)

            elif method == "rank":

                for c in options.columns:
                    tt = table[c]
                    t = zip( tt, range(nrows) )
                    t.sort()
                    for i,n in zip( map(lambda x: x[1], t), range(nrows)):
                        tt[i] = n

            elif method in ("lower-bound", "upper-bound"):

                boundary = float(options.parameters[0])
                del options.parameters[0]
                new_value = float(options.parameters[0])
                del options.parameters[0]

                if method == "upper-bound":
                    for c in options.columns:                
                        for r in range(nrows):
                            if type(table[c][r]) == types.FloatType and \
                                   table[c][r] > boundary:
                                table[c][r] = new_value
                else:
                    for c in options.columns:                
                        for r in range(nrows):
                            if type(table[c][r]) == types.FloatType and \
                                   table[c][r] < boundary:
                                table[c][r] = new_value

            elif method == "fdr":
                pvalues = []
                for c in options.columns: pvalues.extend( table[c] )

                assert max(pvalues) <= 1.0, "pvalues > 1 in table"
                assert min(pvalues) >= 0, "pvalue < 0 in table"

                # convert to str to avoid test for float downstream
                qvalues = map(str, Stats.adjustPValues( pvalues, method = options.fdr_method ))

                x = 0
                for c in options.columns: 
                    table[c] = qvalues[x:x+nrows]
                    x += nrows

            elif method == "normalize-by-table":

                other_table_name = options.parameters[0]
                del options.parameters[0]
                other_fields, other_table  = CSV.ReadTable( open(other_table_name, "r"), with_header = options.has_headers, as_rows = False )

                # convert all values to float
                for c in options.columns:
                    for r in range(nrows):
                        try:
                            other_table[c][r] = float (other_table[c][r] )
                        except ValueError:
                            continue

                ## set 0s to 1 in the other matrix
                for c in options.columns:            
                    for r in range(nrows):
                        if type(table[c][r]) == types.FloatType and \
                               type(other_table[c][r]) == types.FloatType and \
                               other_table[c][r] != 0:
                               table[c][r] /= other_table[c][r]
                        else:
                            table[c][r] = options.missing_value

        ## convert back
        for c in options.columns:
            for r in range(nrows):
                if type(table[c][r]) == types.FloatType:
                    table[c][r] = options.format % table[c][r]

        options.stdout.write( "\t".join(fields) + "\n" )
        if options.sort_rows:
            old2new = {}
            for r in range(nrows):
                old2new[table[0][r]] = r
            for x in options.sort_rows.split(","):
                if x not in old2new: continue
                r = old2new[x]
                options.stdout.write( "\t".join( [ table[c][r] for c in range(ncols) ] ) + "\n")            
        else:
            for r in range(nrows):
                options.stdout.write( "\t".join( [ table[c][r] for c in range(ncols) ] ) + "\n")

    E.Stop()
    



################################################################################
#   Gene prediction pipeline 
#
#   $Id: GenomicIO.py 2781 2009-09-10 11:33:14Z andreas $
#
#   Copyright (C) 2007 Andreas Heger
#
#   This program is free software; you can redistribute it and/or
#   modify it under the terms of the GNU General Public License
#   as published by the Free Software Foundation; either version 2
#   of the License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#################################################################################
"""
GenomicIO.py - Subroutines for working on I/O of large genomic files
====================================================================

:Author: 
:Release: $Id$
:Date: |today|
:Tags: Python

I tried the Biopython parser, but it was too slow for large genomic chunks.
"""

import os, sys, array, string, re, types, tempfile
USAGE="""python IndexedFasta.py [options] name files

Index fasta formatted files to create a database called "name".
"""

import os, sys, array, string, re, types, optparse

global_last_sbjct_token = None
global_last_sbjct_from = None
global_last_sbjct_to = None

global_index = {}

##------------------------------------------------------------
class MySequence(array.array):

    def __init__(self, *args):
        self.mType = args[0]
        array.array.__init__(self, *args)

    def __getslice__( self, *args):
        """return slice as a string."""
        return array.array.__getslice__(self, *args).tostring()

    def __setslice__( self, start, end, sub):
        """set slice start:end from a string sub."""
        return array.array.__setslice__(self, start, end, array.array( self.mType, sub ))

    def __str__(self):
        return self.tostring()

##------------------------------------------------------------
def index_file( filenames, db_name ):
    """index file/files.

    Two new files are create - db_name.fasta and db_name.idx
    """

    if db_name + ".fasta" in filenames:
        raise ValueError( "database (%s.fasta) is part of input set." % db_name)
        
    outfile_index = open( db_name + ".idx", "w")
    outfile_fasta = open( db_name + ".fasta", "w" )

    if type(filenames) == types.StringType:
        filenames = [filenames]

    identifiers = {}
    first = True
    lsequence = 0
    for filename in filenames:

        infile = open( filename, "r")
        
        for line in infile:

            pos = outfile_fasta.tell()        

            if line[0] == "#":  continue
            
            if line[0] == ">" :
                identifier = re.split("\s", line[1:-1])[0]
                if identifier in identifiers:
                    raise ValueError, "%s occurs more than once in %s and %s" % (identifiers[identifier], filename)
                identifiers[identifier] = filename
                
                if not first:
                    outfile_fasta.write("\n")
                    outfile_index.write("%i\n" % lsequence)
                first = False
                    
                outfile_fasta.write( "%s" % line)
                outfile_index.write( "%s\t%i\t%i\t" % (identifier, pos, outfile_fasta.tell()) )
                lsequence = 0
            else:
                s = re.sub( "\s", "", line[:-1] )
                lsequence += len(s)
                outfile_fasta.write( s )
                
    outfile_index.write("%i\n" % lsequence)
    
##------------------------------------------------------------
def _load_index( filename ):
    """load index from file."""

    global global_index
    for line in open(filename, "r"):
        
        (identifier, offset1, offset2, lsequence) = line[:-1].split("\t")

        offset1 = int(offset1)
        offset2 = int(offset2)
        lsequence = int(lsequence)
        
        global_index[identifier] = (offset1, offset2, lsequence)

##------------------------------------------------------------
def index_exists( filename ):
    """check if a certain file has been indexed."""
    return os.path.exists( filename + ".idx" )

##------------------------------------------------------------
def getSequence( db_name,
                 sbjct_token, sbjct_strand,
                 sbjct_from, sbjct_to,
                 as_array = False,
                 forward_coordinates = False):
    """get genomic fragment.
    """

    if not IndexIsLoaded: _load_index( db_name )

    if sbjct_token not in global_index:
        raise KeyError, "%s not in index" % sbjct_token

    is_negative = sbjct_strand in ("-", 0, "0", "-1")
    pos1, pos2, lsequence = global_index[sbjct_token][1]
    infile = open( db_name + ".fasta" )

    if sbjct_to == 0: sbjct_to = lsquence
    if sbjct_from <= 0: sbjct_from = 0

    if is_negative and not forward_coordinates:
        first_pos = pos2 + max( lsequence - sbjct_to, 0 ) 
        last_pos  = pos2 + lsequence - sbjct_from
    else:
        first_pos = pos2 + max(0, sbjct_from)
        last_pos  = pos2 + min(sbjct_to, lsequence)

    infile.seek( first_pos )
    
    p = MySequence( "c" )

    p.fromfile( infile, last_pos - first_pos)

    if is_negative:
        p.reverse()

    if as_array:
        return p
    else:
        return p[:]

def splitFasta( infile, chunk_size, dir="/tmp", pattern=None ):
    """split a fasta file into a subset of files.

    If pattern is not given, random file names are chosen.
    """

    n = 0
    chunk = 0
    
    def __getFilename( chunk ):
        if pattern:
            outname = pattern % chunk
            outfile = os.open( outname, "w" )
        else:
            (outfile, outname) = tempfile.mkstemp( dir=dir )
        return (outfile, outname)

    outfile, outname = __getFilename(chunk)
    filenames = [ outname ]
    noutput = 0
    for line in infile:
        if line[0] == "#": continue
        if line[0] == ">":
            n += 1
            if n > chunk_size:
                os.close(outfile)
                n = 1               
                outfile, outname = __getFilename(chunk)
                filenames.append( outname )
            noutput += 1
            
        os.write( outfile, line )
        
    os.close(outfile)

    if noutput == 0:
        os.remove( outname )
        return []

    return filenames

class IndexedFasta:

    def __init__( self, dbname ):
        self.mDbname = dbname
        self.mIsLoaded = False

    def __getitem__(self, key ):
        return self.getSequence( key, "+", 0, 0, as_array = True )
        
    def loadIndex( self ):

        self.mIndex = {}
        for line in open(self.mDbname + ".idx", "r"):
            
            (identifier, offset1, offset2, lsequence) = line[:-1].split("\t")
            
            offset1 = int(offset1)
            offset2 = int(offset2)
            lsequence = int(lsequence)
            
            self.mIndex[identifier] = (offset1, offset2, lsequence)

        self.mIsLoaded = True

    def getLength( self, sbjct_token ):
        """return sequence length for sbjct_token."""
        if not self.mIsLoaded: self.loadIndex()
        
        return self.mIndex[sbjct_token][2]

    def getContigs( self ):
        """return hash with contig sizes."""
        if not self.mIsLoaded: self.loadIndex()
        return self.mIndex.keys()

    def getContigSizes( self ):
        """return hash with contig sizes."""
        if not self.mIsLoaded: self.loadIndex()
        contig_sizes = {}
        for key, val in self.mIndex.items():
            contig_sizes[key] = val[2]
        return contig_sizes

    def getSequence( self,
                     sbjct_token, sbjct_strand,
                     sbjct_from, sbjct_to,
                     converter = None,
                     as_array = False):
        """get genomic fragment.
        """

        if not self.mIsLoaded: self.loadIndex()

        if sbjct_token not in self.mIndex:
            raise KeyError, "%s not in index" % sbjct_token

        is_negative = sbjct_strand in ("-", 0, "-1", "0")
        pos1, pos2, lsequence = self.mIndex[sbjct_token]
        infile = open( self.mDbname + ".fasta", "r" )

        if sbjct_to == 0: sbjct_to = lsequence
        if sbjct_to > lsequence: sbjct_to = lsequence
        if sbjct_from <= 0: sbjct_from = 0

        if converter:
            first_pos, last_pos = converter( sbjct_from, sbjct_to,
                                             sbjct_strand in ("+", 1, "1"),
                                             lsequence )
        else:
            first_pos, last_pos = sbjct_from, sbjct_to

        if is_negative:
            first_pos, last_pos = lsequence - last_pos, lsequence - first_pos

        first_pos += pos2
        last_pos += pos2
        
        infile.seek( first_pos )

        p = MySequence( "c" )

        p.fromfile( infile, last_pos - first_pos)

        if is_negative:
            p.reverse()            
            p = MySequence("c", string.translate( p[:], string.maketrans("ACGTacgt", "TGCAtgca") ) )

        if as_array:
            return p
        else:
            return p[:]

## converter functions. There are a few of them, but
## I did not want to call functions within functions.
def __one_forward_closed(x, y, c, l):
    x -= 1
    if not c: x, y = l - y, l - x
    return x, y 
def __zero_forward_closed(x, y, c, l):
    y += 1
    if not c: x, y = l - y, l - x
    return x, y 
def __one_both_closed(x, y, c  = None, l = None):
    return x-1, y
def __zero_both_closed(x, y, c = None, l = None):
    return x, y + 1
def __one_forward_open(x, y, c, l):
    x -= 1
    y -= 1
    if not c: x, y = l - y, l - x
    return x, y 
def __zero_forward_open(x, y, c, l):
    if not c: x, y = l - y, l - x
    return x, y 
def __one_both_open(x, y, c  = None, l = None):
    return x-1, y-1
def __zero_both_open(x, y, c = None, l = None):
    return x, y

def getConverter( format ):
    """return a converter function for
    converting various coordinate schemes into
    0-based, both strand, closed-open ranges.

    converter functions have the parameters
    x, y, s, l: with x and y the coordinates of
    a sequence fragment, s the strand (True is positive)
    and l being the length of the contig.
    """

    data = set(format.split("-"))

    if "one" in data:
        if "forward" in data:
            if "closed" in data:
                return __one_forward_closed                
            else:
                return __one_forward_open
        else:
            if "closed" in data:
                return __one_both_closed
            else:
                return __one_both_open
    else:
        if "forward" in data:
            if "closed" in data:
                return __zero_forward_closed
            else:
                return __zero_forward_open
        else:
            if "closed" in data:
                return __zero_both_closed
            else:
                return __zero_both_open                

if __name__ == "__main__":

    import Experiment
    
    parser = E.OptionParser( version = "%prog version: $Id: GenomicIO.py 2781 2009-09-10 11:33:14Z andreas $", usage = globals()["__doc__"])

    (options, args) = Experiment.Start( parser )

    if len(args) < 2:
        print USAGE
        sys.exit(1)

    if options.loglevel >= 1:
        options.stdlog.write("# indexing the following files: %s\n" % (",".join( args[1:] ) ))
        options.stdlog.write("# creating database %s\n" % args[0])
        options.stdlog.flush()
        
    index_file( args[1:], args[0] )
    
    Experiment.Stop()



################################################################################
#   Gene prediction pipeline 
#
#   $Id: Genomics.py 2882 2010-04-07 08:45:52Z andreas $
#
#   Copyright (C) 2004 Andreas Heger
#
#   This program is free software; you can redistribute it and/or
#   modify it under the terms of the GNU General Public License
#   as published by the Free Software Foundation; either version 2
#   of the License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#################################################################################
"""
Genomics.py - Tools for working with genomic data
=================================================

:Author: Andreas Heger
:Release: $Id$
:Date: |today|
:Tags: Python

Code
----

"""

import numpy
import os, sys, string, re, time, array, hashlib, base64, tempfile, math
try: import alignlib
except ImportError: pass

import AString
import Bio.Alphabet

## For caching genome files: remember last one read
global_last_filename_genome = None
global_forward_sequences = {}
global_last_sbjct_token = None

global_translator = string.maketrans("ACGTacgt", "TGCAtgca")

##------------------------------------------------------------
def complement( s ):
    """complement a sequence."""
    return string.translate( s[:], global_translator )[::-1]

##------------------------------------------------------------
def GetHID ( sequence ):
    """returns a hash identifier for a sequence.
    """
    
    # do the encryption
    h = hashlib.md5(sequence).digest()
    # map to printable letters: hid has length 22, so the padded '=' are
    # truncated. You have to add them, if you ever want to decode,
    # but who would do such a thing :=)

    r = base64.encodestring(h)[0:22]
    
    # finally substitute some characters:
    # '/' for '_', so we have legal file names
    # '[' for '+' and ']' for '=' for internet-applications
    
    hid = string.replace(r  , '/', '_')
    hid = string.replace(hid, '+', '[')
    hid = string.replace(hid, '=', ']')
                                                
    return hid

##------------------------------------------------------------
def String2Location( s ):
    """convert a string to location information."""

    data = re.split("[:]+", s)
    if len(data) == 3:
        return data[0], "+", int(data[1]), int(data[2])
    elif len(data) == 4:
        return data[0], data[1], int(data[2]), int(data[3])    
    else:
        raise ValueError, "unknown format %s" % (s)
    
##------------------------------------------------------------
def GetFastaId(fasta_record):
    """Get Fasta Id from a fasta record."""
    title_atoms = string.split(fasta_record.title)
    return title_atoms[0]

##------------------------------------------------------------
def IndexExists( filename ):
    """check if a certain file has been indexed."""
    return os.path.exists( filename + ".idx" )

##------------------------------------------------------------
def IndexFastaFile( filename ):
    Bio.Fasta.index_file(filename, filename + ".idx", GetFastaId )

##------------------------------------------------------------
def ParseFasta2HashFromIndex( filename, filter = None):
    """parse fasta file via index."""
    
    parser = Bio.Fasta.SequenceParser(Bio.Alphabet.IUPAC.ambiguous_dna)

    ## if files and indices have moved, opening the index can be problem,
    ## because file names are relative to the current working directory.
    ## If an absolute file path is given, change first into that directory.
    start_dir = os.getcwd()
    d = os.path.dirname(filename)
    if d: os.chdir( d )
    index = Bio.Fasta.Dictionary(filename + ".idx", parser )
    
    parsed = {}
    
    for id in filter.keys():
        try:
            seq_record = index[id]
        except ZeroDivisionError:
            print "# warning: not found %s" % id
            continue
        except KeyError:
            print "# warning: not found %s" % id
            continue
        parsed[id] = seq_record.seq.tostring()

    os.chdir( start_dir )
        
    return parsed

def ParseFasta2Hash( infile, filter = None, regex_identifier = None):
    """read fasta formatted sequences file and build a hash.

    Keys are all characters before the first whitespace in the description line.

    Previously, if the key contained a ":", everything before the ":" was removed.
    This is not true any more.

    Use array for higher space efficiency.

    If regex_identifier is given, this is used to extract the identifier
    from the fasta description line.
    """
    parsed = {}
    key = None
    p = AString.AString()
    if regex_identifier:
        rx = regex_identifier
    else:
        rx = re.compile( "^(\S+)" )
        
    for line in infile:
        if line[0] == "#": continue
        if not line: continue
        
        if line[0] == ">":
            if key:
                if not filter or key in filter:
                    parsed[key] = p

            key = rx.search( line[1:-1]).groups()[0]
            p = AString.AString()
            continue

        p.extend( AString.AString( re.sub( "\s", "", line[:-1])) )

    if not filter or key in filter:
        if key:
            parsed[key] = p

    return parsed

def oldParseFasta2Hash( infile, filter = None):
    """read fasta formatted sequences file and build a hash.

    Keys are all characters before the first whitespace in the description line.
    If the key contains a :, everything before the : is removed

    Use array for higher space efficiency.
    """
    parsed = {}
    key = None
    fragments = []
    for line in infile:
        if line[0] == "#": continue
        if not line: continue
        
        if line[0] == ">":
            if key:
                if not filter or key in filter:
                    parsed[key] = AString.AString( re.sub( "\s", "", string.join( fragments, "")))
                
            key = re.match(">(\S+)", line[:-1]).groups()[0]
            x = string.find(key,":")
            if x != -1: key = key[x+1:]
            fragments = []
            continue
        
        fragments.append( line[:-1] )

    if not filter or key in filter:
        parsed[key] = AString.AString( re.sub( "\s", "", string.join( fragments, "")))

    return parsed


##------------------------------------------------------------
def WriteGenomicSequences( outfile, forward_sequences, wrap = 0 ):
    """read genomic sequences from a fasta file and return forward
    and backward sequence.
    """
    
    for key, sequence in forward_sequences.items():
        outfile.write(">%s\n" % key)
        if wrap:
            for x in range( 0, len(sequence), wrap):
                outfile.write( "%s\n" % sequence[x:x+wrap] )
        else:
            outfile.write( "%s\n" % sequence[:] )
            
##------------------------------------------------------------
def ReadGenomicSequences( infile,
                          do_reverse = True,
                          filter = None,
                          mask = None,
                          as_array = False ):
    """read genomic sequences from a fasta file and return forward
    and backward sequence.
    """
    
    # read complete genomic sequence
    forward_sequences = ParseFasta2Hash( infile, filter )

    if mask:
        for k in forward_sequences.keys():
            forward_sequences[k] = AString.AString( string.translate( forward_sequences[k][:], string.maketrans("acgtn", "NNNNN")))

    if do_reverse:
        reverse_sequences = {}
        for k, s in forward_sequences.items():
            reverse_sequences[k] = AString.AString( complement( s[:] ) )

    if not as_array:
        for k in forward_sequences.keys():
            forward_sequences[k] = forward_sequences[k][:]
            if do_reverse:
                reverse_sequences[k] = reverse_sequences[k][:]
            
    if do_reverse:
        return forward_sequences, reverse_sequences
    else:
        return forward_sequences

##------------------------------------------------------------
def GetGenomicSequence( sbjct_token, sbjct_strand,
                        sbjct_from, sbjct_to,
                        genome_file,
                        filter = None,
                        use_index = True,
                        loglevel = 2):
    """get genomic fragment."""
    
    global global_last_filename_genome
    global global_forward_sequences
    global global_last_sbjct_token
    
    if "%s" in genome_file:
        filename_genome = genome_file % sbjct_token
    else:
        filename_genome = genome_file

    is_new_file = global_last_filename_genome != filename_genome
    is_new_token = global_last_sbjct_token != sbjct_token
    is_index = use_index and IndexExists( filename_genome )

    ## read new data, if new file is required
    if is_new_file:

        if is_index:
            if loglevel >= 1:
                print "# acquiring %s via index for file %s" % ( sbjct_token, filename_genome)
                sys.stdout.flush()
            
                global_forward_sequences = ParseFasta2HashFromIndex( filename_genome, { sbjct_token: 1 } )

        else:
            if loglevel >= 1:
                print "# reading genome from %s" % filename_genome
                sys.stdout.flush()

            try:
                global_forward_sequences = ReadGenomicSequences( open(filename_genome, "r"),
                                                                 do_reverse = 0,
                                                                 as_array = True,
                                                                 filter = filter)
            except IOError:
                raise "# ERROR: genome %s not found" % filename_genome

    ## read new sbjct_token from via index, if needed
    elif is_new_token:
        
        if is_index:
            if loglevel >= 1:
                print "# acquiring %s via index for file %s" % ( sbjct_token, filename_genome)
                sys.stdout.flush()
            
            global_forward_sequences = ParseFasta2HashFromIndex( filename_genome, { sbjct_token: 1 } )            
    
    if sbjct_token not in global_forward_sequences:
        raise IndexError, "%s not found in %s" % (sbjct_token, filename_genome)
        
    global_last_filename_genome = filename_genome
    global_last_sbjct_token = sbjct_token
    
    if sbjct_strand == "+":
        if sbjct_from and sbjct_to:
            return global_forward_sequences[sbjct_token][sbjct_from:sbjct_to]
        else:
            return global_forward_sequences[sbjct_token]     
    else:
        if sbjct_from and sbjct_to:
            l = len(global_forward_sequences[sbjct_token])
            s = global_forward_sequences[sbjct_token][l-sbjct_to:l-sbjct_from]
            return complement( s ) 
        else:
            return complement( global_forward_sequence[sbjct_token] )

##------------------------------------------------------------
def ReadPeptideSequences( infile, filter = None, as_array = False, regex_identifier = None ):
    """read peptide sequence from fasta infile.
    """

    sequences = ParseFasta2Hash(infile, filter, regex_identifier = regex_identifier )

    if not as_array: 
        for k in sequences.keys():
            sequences[k] = sequences[k][:]
    return sequences

##------------------------------------------------------------
def ReadContigSizes( infile ):
    """read sizes of congits from file.
    """
    sizes = {}
    for line in infile:
        if line[0] == "#":continue
        sbjct_token, size = string.split( line[:-1], "\t")[:2]
        sizes[sbjct_token] = int(size)

    return sizes

##------------------------------------------------------------
def ReadLocationsGFF( infile ):
    """read locations from a file in GFF format
    """

    class Location :
        def __init__( self ): pass
        def __str__(self):
            return string.join(map(str, (self.mQueryToken, self.mSbjctGenomeFrom, self.mSbjctGenomeTo,
                                         self.mSbjctStrand, self.mSbjctToken)), "\t")
                
    locations = {}
    for line in infile:
        if line[0] == "#": continue
        l = Location()
        l.mQueryToken, l.mSbjctGenomeFrom, l.mSbjctGenomeTo, l.mSbjctStrand, l.mSbjctToken = line[:-1].split("\t")
        l.mSbjctGenomeFrom, l.mSbjctGenomeTo = map(int, (l.mSbjctGenomeFrom, l.mSbjctGenomeTo ))
        locations[l.mQueryToken] = l

    return locations

##------------------------------------------------------------
def ReadGo( infile ):
    """read go assignments from file as output by Caleb's script
    """

    class Go: 
        def __init__( self ): pass
        def __str__(self):
            return string.join(map(str, (self.mGoid, self.mDescription, self.mEvidence)), "\t")
        
    go = {}
    for line in infile:
        if line[0] == "#": continue
        g = Go()
        id, g.mGoid, g.mDescription, g.mEvidence = line[:-1].split(" : ")
        if id not in go: go[id] = []
        go[id].append( g )

    return go
    
##------------------------------------------------------------
def ReadClusters( infile ):
    """read clustering information from cd-hit file.
    """

    rep = None
    map_rep2mem = {}
    map_mem2rep = {}
    mems = None
    for line in infile:
        if line[:8] == ">Cluster":
            if rep:
                for mem in mems: map_mem2rep[mem] = rep
                map_rep2mem[rep] = mems
            mems = []
            rep = None
            continue
        
        try:
            (n, length, id, level) = re.search("(\d)\s+(\d+)aa,\s+>([\S\s]+)\s+(.*)", line[:-1]).groups()
            id = re.split("\s", id)[0]
            if id[-3:] == "...": id = id[:-3]
        except AttributeError:
            raise "parsing error in line: %s" % line[:-1]
        
        if level == "*":
            rep = id
        else:
            mems.append(id)
        
    if rep:
        for mem in mems: map_mem2rep[mem] = rep
        map_rep2mem[rep] = mems

    return map_rep2mem, map_mem2rep

##------------------------------------------------------------
def ToForwardCoordinates( a, b, strand, length ):
    """return a and b in forward coordinates."""
    if strand in ("-", "0", "-1", 0):
        return length - b, length - a
    else:
        return a, b

##------------------------------------------------------------
def ReadMap( infile ):
    """read clustering information from a map file.

    The file maps transcripts to representative transcripts.
    rep is not a member of itself.
    """

    rep = None
    map_rep2mem = {}
    map_mem2rep = {}
    mems = None
    for line in infile:

        if line[0] == "#": continue
        rep, mem = line[:-1].split("\t")[:2]

        if rep == mem: continue
        
        if rep not in map_rep2mem: map_rep2mem[rep] = []

        map_rep2mem[rep].append( mem )
        map_mem2rep[mem] = rep

    return map_rep2mem, map_mem2rep

##------------------------------------------------------------
def CountGeneFeatures( first_position,
                       alignment,
                       genomic_sequence = None,
                       border_stop_codon = 0,
                       stop_codons = ("TAG", "TAA", "TGA") ):
    """calculate number of genomic features in a peptide to genome alignment.

    border_stop_codon should be divisible by three and is the number of codons
    that are ignored at the edes of matches regions.

    returns (nintrons, nframeshifts, ngaps, nsplit, nstopcodons, disruptions)

    disruptions is a list of disruptions in the prediction. Each disruption
    is a tuple of ( "stop|frameshift", position in protein, position in cds, position on genomic sequence).

    Note that codons can be split, for example:
    S 0 2 5 0 2 I 0 17541 3 0 2 S 1 2 5 0 2 I 0 27979 3 0 2 S 1 2
    """

    current_pos_genome = first_position
    current_pos_cds = 0
    
    nstopcodons = 0
    nintrons = 0
    nframeshifts = 0
    ngaps = 0
    nsplits = 0

    disruptions = []

    # position of nucleotides for split codons
    partial_codon = []
    
    first_state = True

    for state, l_protein, l_genome in alignment:

        if state in ("G", "M", "P"):

            ## reset split codon. Do a sanity check, but ignore first split codon
            # in the case of exon alignments
            if len(partial_codon) % 3 != 0 and not first_state:
                raise ValueError ("split codon was not multiple of three: partial_codon=%s, alignment=%s" % \
                                      (partial_codon, alignment ) )
                # print alignment
                # pass

            partial_codon = []
            first_state = False
            
        if state in ("M", "G"):
            # check for stop codons, ignore the first border_stop_codon nucleotides
            # note: this should be a multiple of three
            if genomic_sequence:
                
                for x in range(  border_stop_codon, l_genome - border_stop_codon, 3):
                    y = current_pos_cds + x
                    z = current_pos_genome + x
                    codon = genomic_sequence[z:z+3].upper()
                    
                    if codon in stop_codons:
                        nstopcodons += 1
                        disruptions.append( ( "stop",
                                              y, y+3,
                                              z, z+3) )
                        
        if state == "I":
            nintrons += 1
            
        elif state == "F":
            nframeshifts += 1
            disruptions.append( ( "frameshift",
                                  current_pos_cds, current_pos_cds + l_genome,
                                  current_pos_genome, current_pos_genome + l_genome ) )
            
        elif state == "G":
            ngaps += 1
            
        elif state == "P":
            ngaps += 1

        elif state == "S":

            # I used to ignore alignments that start with a split codon as they
            # might have been incomplet. However, the code below seems to work for split codens
            # if not first_state:
            nsplits += 1

            ## check for stop-codons in split codons as well:
            for x in range(l_genome):
                partial_codon.append( (current_pos_cds + x, current_pos_genome + x) )
                
            if border_stop_codon < 3 and genomic_sequence and len(partial_codon) % 3 == 0:

                for x in range( 0, len(partial_codon), 3 ):
                    codon = "".join(genomic_sequence[c[1]] for c in partial_codon[x:x+3]).upper()
                    if codon in stop_codons:
                        nstopcodons += 1
                        disruptions.append( ("split-stop",
                                             partial_codon[x][0], partial_codon[x+2][0],
                                             partial_codon[x][1], partial_codon[x+2][1]) )
        ## advance in cds
        if state in ( "M", "G", "F", "S"):
            current_pos_cds += l_genome
            
        current_pos_genome += l_genome

    return nintrons, nframeshifts, ngaps, nsplits, nstopcodons, disruptions

##------------------------------------------------------------
def Alignment2String( alignment ):
    """convert an alignment to an alignment string.
    """
    return string.join( map( \
                       lambda x: string.join(map(str, x), " "),
                       alignment), " ")

##------------------------------------------------------------
def String2Alignment( source ):
    """convert an alignment string to an alignment.
    """

    d = string.split( source, " ")
    ali = []
    if len(d) < 3:
        return ali

    for x in range(0, len(d), 3):
        ali.append( (d[x], int(d[x+1]), int(d[x+2]) ) )
        
    return ali

##------------------------------------------------------------
def GetAlignmentLength( alignment ):

    q, s = 0, 0
    for state, l_query, l_sbjct in alignment:
        q += l_query
        s += l_sbjct

    return q, s

##------------------------------------------------------------
def Alignment2ExonBoundaries( alignment, query_from = 0, sbjct_from = 0, add_stop_codon = 1):
    """convert a Peptide2DNA alignment to exon boundaries.

    frame is frame of start of exon.
    """

    exons = []

    ## count in nucleotides for query
    query_from *= 3
    query_pos = query_from
    sbjct_pos = sbjct_from
    frame = 0
    ali = []

    p_as_intron = True
    if p_as_intron:
        ali_states = ("S", "M", "G", "F" )
    else:
        ali_states = ("S", "M", "G", "F", "P" )
        
    for state, l_query, l_sbjct in alignment:

        ## count as nucleotides
        l_query *= 3

        if state in ali_states:
            ali.append( (state, l_query, l_sbjct))
                
        if state == "M":
            query_pos += l_query

        elif state == "G":
            query_pos += l_query

        elif state == "P":
            ## treat as intron
            if p_as_intron:
                frame = query_from % 3
                if frame != 0: frame = 3 - frame
                exons.append( (query_from, query_pos, frame, sbjct_from, sbjct_pos, ali) )
                ali = []
                query_from = query_pos
                sbjct_from = sbjct_pos + l_sbjct
                
            query_pos += l_query
            
        elif state == "S":
            query_pos += l_sbjct

        elif state == "I":
            pass
            
        elif state == "5":
            ## frame is frame for this exon
            frame = query_from % 3
            if frame != 0: frame = 3 - frame
            exons.append( (query_from, query_pos, frame, sbjct_from, sbjct_pos, ali) )
            ali = []
            query_pos += l_query
            
        elif state == "3":
            query_pos += l_query
            query_from = query_pos
            sbjct_from = sbjct_pos + l_sbjct
            
        sbjct_pos += l_sbjct

                
    ## add three for the stop codon:
    if add_stop_codon:
        query_pos += 3

    frame = query_from % 3
    if frame != 0: frame = 3 - frame

    exons.append( (query_from, query_pos, frame, sbjct_from, sbjct_pos, ali) )

    return exons

##------------------------------------------------------------
def RemoveFrameShiftsFromAlignment( row_ali, col_ali, gap_char = "-" ):
    """remove frame shifts in alignment. These are all gaps that
    are 1, 2, 4, or 5 residues long."""


    match_string = "[^%s]%s+[^%s]" %\
                       (gap_char, gap_char, gap_char)

    positions = []
    for x in re.finditer( match_string, row_ali):
        positions.append( (x.start()+1, x.end()-1) )
        
    for x in re.finditer( match_string, col_ali):
        positions.append( (x.start()+1, x.end()-1) )

    # cut and paste new alignment
    new_row_ali = ""
    new_col_ali = ""
    a = 0
    for first,last in positions:
        if (last - first) % 3:
            new_row_ali += row_ali[a:first]
            new_col_ali += col_ali[a:first]        
            a = last
        
    new_row_ali += row_ali[a:]
    new_col_ali += col_ali[a:]    
        
    return new_row_ali, new_col_ali

##------------------------------------------------------------
def IsStopCodon( codon, stop_codons = ("TAG", "TAA", "TGA")):
    return codon in stop_codons
                        
##------------------------------------------------------------
def MaskStopCodons( sequence, stop_codons = ("TAG", "TAA", "TGA")):
    """mask stop codons in a sequence.
    """
    codons = []
    
    for x in range(0, len(sequence),3):
        codon = sequence[x:x+3]
        if codon in stop_codons: codon = "NNN"
        codons.append( codon )
    return string.join( codons, "")

##------------------------------------------------------------
def Alignment2DNA( alignment, query_from = 0, sbjct_from = 0):
    """convert a Peptide2DNA alignment to a DNA against DNA
    alignment.
    """

    map_query2sbjct = alignlib.makeAlignmentVector()

    ## count in nucleotides for query
    query_pos = query_from * 3
    sbjct_pos = sbjct_from

    for state, l_query, l_sbjct in alignment:

        ## count as nucleotides
        l_query *= 3
        
        if state in ("A", "B", "C"):
            
            if state in ("A"):
                l_query = 0
            elif state in ("B"):
                l_query = 1
            elif state in ("C"):
                l_query = 2 
                
        elif state in ("a", "b", "c"):

            if state in ("a"):
                l_query = 0
            elif state in ("b"):
                l_query = 2
            elif state in ("c"):
                l_query = 1

        elif state == "S":
            l_query = l_sbjct
                
        if l_query > 0 and l_sbjct > 0:
            alignlib.addDiagonal2Alignment( map_query2sbjct,
                                            query_pos, query_pos + l_query,
                                            sbjct_pos - query_pos )
                
        query_pos += l_query
        sbjct_pos += l_sbjct

    return map_query2sbjct



# W: ACGT
# Y: CT = Pyrimidine
# R: AG = Purine

Wobble = {
    "A" : "GCW",
    "C" : "TGY",
    "D" : "GAY",
    "E" : "GAR",
    "F" : "TTY",
    "G" : "GGW",
    "H" : "CAY",
    "I" : "ATW",
    "K" : "AAR",
    "L" : "WTW",
    "M" : "ATG",
    "N" : "AAY",
    "P" : "CCW",
    "Q" : "CAR",
    "R" : "WGW",
    "S" : "WWW",
    "T" : "ACW",
    'U' : "TGA",        # selenocysteine
    "V" : "GTW",
    "W" : "TGG",
    "X" : "WWW",        # X can be stop codon or masked sequence.
    "Y" : "TAW",
    "*" : "NNN",
    "?" : "WWW",        # ? can be stop codon or masked sequence.
    "-" : "---",
    }

GeneticCode = {
    "TTT" : "F",
    "TTC" : "F",
    "TTA" : "L",
    "TTG" : "L",
    "TCT" : "S",
    "TCC" : "S",
    "TCA" : "S",
    "TCG" : "S",
    "TAT" : "Y",
    "TAC" : "Y",
    "TAA" : "X", # stop
    "TAG" : "X", # stop
    "TGT" : "C",
    "TGC" : "C",
    "TGA" : "X", # stop
    "TGG" : "W",
    "CTT" : "L",
    "CTC" : "L",
    "CTA" : "L",
    "CTG" : "L",
    "CCT" : "P",
    "CCC" : "P",
    "CCA" : "P",
    "CCG" : "P",
    "CAT" : "H",
    "CAC" : "H",
    "CAA" : "Q",
    "CAG" : "Q",
    "CGT" : "R",
    "CGC" : "R",
    "CGA" : "R",
    "CGG" : "R",
    "ATT" : "I",
    "ATC" : "I",
    "ATA" : "I",
    "ATG" : "M",
    "ACT" : "T",
    "ACC" : "T",
    "ACA" : "T",
    "ACG" : "T",
    "AAT" : "N",
    "AAC" : "N",
    "AAA" : "K",
    "AAG" : "K",
    "AGT" : "S",
    "AGC" : "S",
    "AGA" : "R",
    "AGG" : "R",
    "GTT" : "V",
    "GTC" : "V",
    "GTA" : "V",
    "GTG" : "V",
    "GCT" : "A",
    "GCC" : "A",
    "GCA" : "A",
    "GCG" : "A",
    "GAT" : "D",
    "GAC" : "D",
    "GAA" : "E",
    "GAG" : "E",
    "GGT" : "G",
    "GGC" : "G",
    "GGA" : "G",
    "GGG" : "G",
    }

GeneticCodeSeleno = GeneticCode.copy()
GeneticCodeSeleno["TGA"] = "U"

GeneticCodeAA = {
    "TTT" : "F",
    "TTC" : "F",
    "TTA" : "L",
    "TTG" : "L",
    "TCT" : "S",
    "TCC" : "S",
    "TCA" : "S",
    "TCG" : "S",
    "TAT" : "Y",
    "TAC" : "Y",
    "TGT" : "C",
    "TGC" : "C",
    "TGG" : "W",
    "CTT" : "L",
    "CTC" : "L",
    "CTA" : "L",
    "CTG" : "L",
    "CCT" : "P",
    "CCC" : "P",
    "CCA" : "P",
    "CCG" : "P",
    "CAT" : "H",
    "CAC" : "H",
    "CAA" : "Q",
    "CAG" : "Q",
    "CGT" : "R",
    "CGC" : "R",
    "CGA" : "R",
    "CGG" : "R",
    "ATT" : "I",
    "ATC" : "I",
    "ATA" : "I",
    "ATG" : "M",
    "ACT" : "T",
    "ACC" : "T",
    "ACA" : "T",
    "ACG" : "T",
    "AAT" : "N",
    "AAC" : "N",
    "AAA" : "K",
    "AAG" : "K",
    "AGT" : "S",
    "AGC" : "S",
    "AGA" : "R",
    "AGG" : "R",
    "GTT" : "V",
    "GTC" : "V",
    "GTA" : "V",
    "GTG" : "V",
    "GCT" : "A",
    "GCC" : "A",
    "GCA" : "A",
    "GCG" : "A",
    "GAT" : "D",
    "GAC" : "D",
    "GAA" : "E",
    "GAG" : "E",
    "GGT" : "G",
    "GGC" : "G",
    "GGA" : "G",
    "GGG" : "G",
    }

StopCodons = ("TAG", "TAA", "TGA")

## degenerate mappings of the first two nucleotides
## to amino acid
GeneticCodeDegenerate = {
    "TC" : "S",
    "CT" : "L",
    "CC" : "P",
    "CG" : "R",
    "AC" : "T",
    "GT" : "V",
    "GC" : "A",
    "GG" : "G",
    }


## first and third position done, second position not checked
## Ile is two-fold, even there are three positions
Degeneracy = {
    "TTT" : ("F",1,1,2),
    "TTC" : ("F",1,1,2), 
    "TTA" : ("L",2,1,2),
    "TTG" : ("L",2,1,2),
    "TCT" : ("S",1,1,4),
    "TCC" : ("S",1,1,4),
    "TCA" : ("S",1,1,4),
    "TCG" : ("S",1,1,4),
    "TAT" : ("Y",1,1,2),
    "TAC" : ("Y",1,1,2),
    "TGT" : ("C",1,1,2),
    "TGC" : ("C",1,1,2),
    "TGG" : ("W",1,1,1),
    "CTT" : ("L",1,1,4),
    "CTC" : ("L",1,1,4),
    "CTA" : ("L",2,1,4),
    "CTG" : ("L",2,1,4),
    "CCT" : ("P",1,1,4),
    "CCC" : ("P",1,1,4),
    "CCA" : ("P",1,1,4),
    "CCG" : ("P",1,1,4),
    "CAT" : ("H",1,1,2),
    "CAC" : ("H",1,1,2),
    "CAA" : ("Q",1,1,2),
    "CAG" : ("Q",1,1,2),
    "CGT" : ("R",1,1,4),
    "CGC" : ("R",1,1,4),
    "CGA" : ("R",2,1,4),
    "CGG" : ("R",2,1,4),
    "ATT" : ("I",1,1,3),
    "ATC" : ("I",1,1,3),
    "ATA" : ("I",1,1,3),
    "ATG" : ("M",1,1,1),
    "ACT" : ("T",1,1,4),
    "ACC" : ("T",1,1,4),
    "ACA" : ("T",1,1,4),
    "ACG" : ("T",1,1,4),
    "AAT" : ("N",1,1,2),
    "AAC" : ("N",1,1,2),
    "AAA" : ("K",1,1,2),
    "AAG" : ("K",1,1,2),
    "AGT" : ("S",1,1,2),
    "AGC" : ("S",1,1,2),
    "AGA" : ("R",2,1,2),
    "AGG" : ("R",2,1,2),
    "GTT" : ("V",1,1,4),
    "GTC" : ("V",1,1,4),
    "GTA" : ("V",1,1,4),
    "GTG" : ("V",1,1,4),
    "GCT" : ("A",1,1,4),
    "GCC" : ("A",1,1,4),
    "GCA" : ("A",1,1,4),
    "GCG" : ("A",1,1,4),
    "GAT" : ("D",1,1,2),
    "GAC" : ("D",1,1,2),
    "GAA" : ("E",1,1,2),
    "GAG" : ("E",1,1,2),
    "GGT" : ("G",1,1,4),
    "GGC" : ("G",1,1,4),
    "GGA" : ("G",1,1,4),
    "GGG" : ("G",1,1,4),
    }

GAP_CHAR="-"

## amino acid based degeneracy
DegeneracyAA = {
    "F" : 2,
    "L" : 6,
    "S" : 4,
    "Y" : 2,
    "C" : 2,
    "W" : 1,
    "P" : 4,
    "H" : 2,
    "Q" : 2,
    "R" : 4,
    "I" : 3,
    "M" : 1,
    "T" : 4,
    "N" : 2,
    "K" : 2,
    "S" : 2,
    "R" : 2,
    "V" : 4,
    "A" : 4,
    "D" : 2,
    "E" : 2,
    "G" : 4,
    "X" : 0,
    "-" : 0,
    "." : 0,
    }

# map ambiguous nucleic acid codes to the four bases
AMBIGUOUS_CODES_NA = {
    'A': 'A',
    'C': 'C',
    'G': 'G',
    'T': 'T',
    'N': 'ACGT',
    'X': 'ACGT',
    'U': 'T',
    'R': 'AG',
    'Y': 'CT',
    'M': 'AC',
    'K': 'GT',
    'S': 'CG',
    'W': 'AT',
    'H': 'ACT',
    'B': 'CGT',
    'V': 'ACG',
    'D': 'AGT' }

REVERSE_AMBIGUOUS_NA = {
    'A': 'A', 'C': 'C', 'G': 'G', 'T': 'T',
    'AA': 'A', 'CC': 'C', 'GG': 'G', 'TT': 'T', 'UU': 'U',
    'AG': 'R', 'GA': 'R',
    'CT': 'Y', 'TC': 'Y',
    'AC': 'M', 'CA': 'M',
    'GT': 'K', 'TG': 'K',
    'CG': 'S', 'GC': 'S',
    'AT': 'W', 'TA': 'W',
    'ACT': 'H', 'ATC': 'H', 'CAT': 'H', 'CTA': 'H', 'TAC': 'H', 'TCA': 'H', 
    'CGT': 'B', 'CTG': 'B', 'GCT': 'B', 'GTC': 'B', 'TCG': 'B', 'TGC': 'B',
    'ACG': 'V', 'AGC': 'V', 'CAG': 'V', 'CGA': 'V', 'GAC': 'V', 'GCA': 'V',
    'AGT': 'D', 'ATG': 'D', 'GAT': 'D', 'GTA': 'D', 'TAG': 'D', 'TGA': 'D', 
    }        

ENCODE_GENOTYPE = {
    'A': 'A', 'C': 'C', 'G': 'G', 'T': 'T',
    'AA': 'A', 'CC': 'C', 'GG': 'G', 'TT': 'T', 'UU': 'U',
    'AG': 'r', 'GA': 'R',
    'CT': 'y', 'TC': 'Y',
    'AC': 'm', 'CA': 'M',
    'GT': 'k', 'TG': 'K',
    'CG': 's', 'GC': 'S',
    'AT': 'w', 'TA': 'W',
    }        

DECODE_GENOTYPE = {
    'A': 'AA',
    'C': 'CC',
    'G': 'GG',
    'T': 'TT',
    'r': 'AG', 'R': 'AG',
    'y': 'CT', 'Y': 'CT',
    'm': 'AC', 'M': 'AC',
    'k': 'GT', 'K': 'GT',
    's': 'CG', 'S': 'CG',
    'w': 'AT', 'W': 'AT',
    }

##------------------------------------------------------------
def encodeGenotype( code ):
    '''encode genotypes like GG, GA into a one-letter code.
    The returned code is lower case if code[0] < code[1], otherwise
    it is uppercase.
    '''
    return ENCODE_GENOTYPE[ code.upper() ]

def decodeGenotype( code ):
    '''decode single letter genotypes like m, M into two letters.
    This is the reverse operation to :meth:`encodeGenotype`.
    '''
    return DECODE_GENOTYPE[ code ] 

##------------------------------------------------------------
def resolveAmbiguousNA( code ):
    '''resolve ambiguous nucleic acid letters.
    '''
    return AMBIGUOUS_CODES_NA[ code.upper() ]

##------------------------------------------------------------
def resolveReverseAmbiguousNA( genotype ):
    '''map a genotype to a single letter amino acid amiguous code,
    for example, CT -> Y.
    '''
    return REVERSE_AMBIGUOUS_NA[ genotype.upper() ]

##------------------------------------------------------------
def GetMapAA2Codons():
    """returns a map of amino acids to codons

    No stop codons.
    ."""
    map_aa2codons = {}

    for codon,aa in GeneticCodeAA.items():
        if aa not in map_aa2codons: map_aa2codons[aa] = []
        map_aa2codons[aa].append( codon )
    return map_aa2codons

##------------------------------------------------------------
def GetDegeneracy( codon ):
    return Degeneracy[codon.upper()]

##------------------------------------------------------------
def IsStopCodon( codon, stop_codons = ("TAG", "TAA", "TGA")):
    return codon in stop_codons

##------------------------------------------------------------
def MapCodon2AA( codon, is_seleno = False, ignore_n = True ):
    '''map a codon to an amino acid using the standard translation
    tables

    The mapping returns gaps as gaps and will return an amino acid
    for incomplete codons if there is unambiguous mapping.

    If ``is_seleno`` is set, the codon is translated for a selenoprotein.

    If ``ignore_n`` is set, codons with ``n`` are returned
    as ``?`` in order to distinguish them from stop codons.

    Amino acids are returned as upper-case letters.
    '''

    codon = string.upper(codon)

    codon = re.sub( "[.-]", "", codon)
    if len(codon) == 0: return GAP_CHAR

    if is_seleno: 
        code = GeneticCodeSeleno
    else:
        code = GeneticCode
    
    if code.has_key(codon):
        return code[codon]
    elif codon == "---":
        return "-"
    elif GeneticCodeDegenerate.has_key(codon[:2]):
        ## check for four-fold degenerate codons,
        ## if they can be mapped (ENSEMBL does it).
        return GeneticCodeDegenerate[codon[:2]]        
    elif ignore_n and ("n" in codon or "N" in codon):
        return "?"
    else:
        return "X"

##------------------------------------------------------------
def Protein2Wobble( s  ):

    c = []
    for x in s:
        c.append( Wobble[x] )
    return string.join(c, "")

##------------------------------------------------------------
def Alignment2PeptideAlignment( alignment,
                                query_from = 0,
                                sbjct_from = 0,
                                genomic_sequence = None):
    """convert a Peptide2DNA aligment to a Peptide2Peptide alignment.

    How to handle frameshifts?
    """

    map_query2sbjct = alignlib.makeAlignmentVector()

    query_pos = query_from
    sbjct_pos = 0
    sbjct_genome_pos = sbjct_from 
    sbjct_residues = []
    codon = ""

    for state, l_query, l_sbjct in alignment:

        query_increment = 0
        sbjct_increment = 0

        if state == "M":
            
            query_increment = l_query
            sbjct_increment = l_sbjct / 3
            if genomic_sequence:
                codon = genomic_sequence[sbjct_genome_pos:sbjct_genome_pos+l_sbjct]
                
        elif state == "S":
            if l_query:
                sbjct_increment = 1
                query_increment = 1
                
            if genomic_sequence:
                codon += genomic_sequence[sbjct_genome_pos:sbjct_genome_pos+l_sbjct]
                
        elif state == "G":
            query_increment = l_query
            sbjct_increment = l_sbjct / 3
            if genomic_sequence:
                codon += genomic_sequence[sbjct_genome_pos:sbjct_genome_pos+l_sbjct]
            
        elif state == "P":
            ## only increment query, sbjct does not advance.
            query_increment = l_query
            
        if query_increment and sbjct_increment:
            alignlib.addDiagonal2Alignment( map_query2sbjct,
                                            query_pos, query_pos + query_increment,
                                            sbjct_pos - query_pos )

        if sbjct_increment and genomic_sequence:
            for x in range(0, len(codon), 3):
                sbjct_residues.append( MapCodon2AA(codon[x:x+3]) )
            codon = ""

        query_pos += query_increment
        sbjct_pos += sbjct_increment
        
        sbjct_genome_pos += l_sbjct

    return map_query2sbjct, string.join( sbjct_residues, "")

##-------------------------------------------------------------------------
def translate( sequence, 
               is_seleno = False,
               prefer_lowercase = True,
               ignore_n = False,
               ):
    '''convert DNA sequence to a peptide sequence

    If ``is_seleno`` is set, "TGA" codons are treated as
    encoding for selenocysteine.

    If ``ignore_n`` is set, codons with ``n`` are returned
    as ``?`` in order to distinguish them from stop codons.

    '''
    residues = []

    for x in range(0, len(sequence), 3):
        if prefer_lowercase:
            is_lower = False
            for c in sequence[x:x+3]: is_lower = is_lower or c in string.lowercase
        else:
            is_lower = True
            for c in sequence[x:x+3]: is_lower = is_lower and c in string.lowercase

        if is_lower:
            residues.append( MapCodon2AA( sequence[x:x+3].upper(), 
                                          is_seleno = is_seleno,
                                          ignore_n = ignore_n ).lower() )
        else:
            residues.append( MapCodon2AA( sequence[x:x+3].upper(), 
                                          is_seleno = is_seleno,
                                          ignore_n = ignore_n ).upper() )

    return string.join( residues, "")

##-------------------------------------------------------------------------
def TranslateDNA2Protein( *args, **kwargs ):
    """convert a DNA sequence to a peptide sequence.
    keep case.

    deprecated - use :meth:`translate` instead.
    """
    return translate( *args, **kwargs )

##-------------------------------------------------------------------------
def Alignment2CDNA( alignment,
                    query_from = 0,
                    sbjct_from = 0,
                    genome = None,                    
                    remove_frameshifts = 0):
    """build cDNA sequence from genomic fragment and
    return alignment of query to it.
    """

    fragments = []
    sbjct_pos = 0
    map_query2sbjct = alignlib.makeAlignmentVector()

    ## count in nucleotides for query
    query_pos = query_from * 3
    sbjct_pos = sbjct_from
    ## position in cDNA
    cdna_pos  = 0
    for state, l_query, l_sbjct in alignment:

        ## count as nucleotides
        l_query *= 3

        keep = False
        
        if state == "M":
            keep = True
        elif state == "S":
            l_query = l_sbjct            
            keep = True
        elif state == "F" and not remove_frameshifts:
            keep = True
        elif state == "G":
            if l_sbjct > 0:
                keep = True
        elif state == "P":
            keep = False

        if keep:
            if genome:
                fragments.append(genome[sbjct_pos:sbjct_pos+l_sbjct])

            if l_query > 0 and l_sbjct > 0:
                alignlib.addDiagonal2Alignment( map_query2sbjct,
                                                query_pos, 
                                                query_pos + l_query,
                                                cdna_pos - query_pos )
            cdna_pos += l_sbjct
                
        query_pos += l_query
        sbjct_pos += l_sbjct

    return map_query2sbjct, string.join(fragments, "")

##-------------------------------------------------------------------
def GetExon( exons, first_aa ):

    first_na = (first_aa - 1) * 3
    combined_na = max(exons[0].mPeptideFrom, first_na)

    ## add not covered positions
    g = exons[0].mGenomeFrom + max(0, combined_na - exons[0].mPeptideFrom)

    ## add introns
    e = 0
    while exons[e].mPeptideTo < combined_na:
        e += 1
        g += exons[e].mGenomeFrom - exons[e-1].mGenomeTo
        
    return g, e, combined_na / 3 + 1

##-------------------------------------------------------------------
def Exons2Alignment( exons ):
    """build an cigar alignment string from a list of exons.
    """

    exons.sort()

    phase = 0
    alignment = []
    last_to = 0
    nexon = 0
    
    for this_from, this_to in exons:

        nexon += 1

        # print alignment        
        # print this_from, this_to

        lintron = this_from - last_to
        lexon = this_to - this_from

        if lintron < 0:
            raise ValueError ("error: invalid intron length: %i" % lintron)

        if last_to:

            if ( phase != 0):
                missing = 3 - phase

                alignment.append( ("S", 0, phase) )

            if lintron >= 4:
                # deal with real introns:
                # at least 4 nucleotides for splice signal
                alignment.append( ("5", 0, 2) )
                alignment.append( ("I", 0, lintron - 4) )
                alignment.append( ("3", 0, 2) )
                
            else:
                ## if phase is 0, regard as frameshift
                if phase == 0:
                    alignment.append( ("F", 0, lintron) )
                else:
                    # deal with frameshift introns
                    # add dummy entries for splice signals
                    alignment.append( ("5", 0, 0) )
                    alignment.append( ("I", 0, lintron) )
                    alignment.append( ("3", 0, 0) )

            if phase != 0:
                ## deal with very small exons

                missing = 3 - phase

                # deal with small exons that can not
                # accomodate a full codon
                if missing > lexon: 
                    alignment.append( ("S", 1, missing) )
                else:
                    alignment.append( ("S", 1, missing) )
                    phase = 0
                    this_from += missing
                
                # raise ValueError ("expecting a split codon, but small exon %i-%i can not accomodate phase %i in exons %s " % (this_from, this_to, phase, str(exons)))
                
        l = this_to - this_from

        if l < 0:
            # note: sometimes the last exon just contains a split codon and the stop codon.
            # Thus: do not raise an error if it is the last exon
            if nexon == len(exons):
                break
            else:
                pass
                # raise ValueError ("error: negative length for aligned residues at exon %i-%i in exons %s" % (this_from, this_to, str(exons)))

        phase = l % 3
        
        ncodons = (l - phase) / 3
        alignment.append( ("M", ncodons, ncodons * 3) )

        last_to = this_to

    return alignment

##-------------------------------------------------------------------
def AlignmentProtein2CDNA( src, exons1 = None, exons2 = None):
    """convert a peptide alignment to a nucleotide
    alignment.

    multiplies coordinates with 3.
    Insert introns.

    Note: alignment starts at 1
    """
    result = src.getNew()

    ## enter correct exons (for example, if alignment is not fully covering or
    ## gene structure is incomplete)
    gx, exon_id1, first_x = GetExon( exons1, src.getRowFrom() )
    gy, exon_id2, first_y = GetExon( exons2, src.getColFrom() )    

    # print "x=", gx, exon_id1, first_x
    # print "y=", gy, exon_id2, first_y
    
    last_x = first_x - 1
    last_y = first_y - 1

    exon1_to = exons1[exon_id1].mGenomeTo
    exon2_to = exons2[exon_id2].mGenomeTo
    
    # only iterate in alignment
    for x in range( first_x, 
                    min(src.getRowTo() + 1, exons1[-1].mPeptideTo / 3)):

        y = src.mapRowToCol(x)
        if not y: continue
        if y < first_y: continue

        gx += (x - last_x - 1) * 3
        gy += (y - last_y - 1) * 3

        # print "# x=", x, "last_x=", last_x, "y=", y, "last_y=", last_y
        
        for i in range( 0, 3):

            # print i, "\t", gx, exons1[exon_id1].mGenomeTo, exon_id2, "\t", gy, exon_id2, exons2[exon_id2].mGenomeTo
            
            if gx >= exon1_to:
                exon_id1 += 1
                if exon_id1 < len(exons1):
                    gx += exons1[exon_id1].mGenomeFrom - exon1_to
                    exon1_to = exons1[exon_id1].mGenomeTo
                    
            if gy >= exon2_to:
                exon_id2 += 1
                if exon_id2 < len(exons2):
                    gy += exons2[exon_id2].mGenomeFrom - exon2_to
                    exon2_to = exons2[exon_id2].mGenomeTo 

            gx += 1
            gy += 1
            result.addPairExplicit( gx, gy, 0 )

        last_x = x
        last_y = y

    return result
    
##-------------------------------------------------------------------
def GetDegenerateSites( seq1, seq2,
                        degeneracy = 4,
                        position = 3 ):
    """returns two new sequenes containing only degenerate sites.

    Only unmutated positions are counted.
    """

    new_seq1 = []
    new_seq2 = []
    for x in range( 0, len(seq1), 3):

        c1 = seq1[x:x+3]
        c2 = seq2[x:x+3]

        if c1 in GeneticCodeAA and c2 in GeneticCodeAA:

            if GeneticCodeAA[c1] == GeneticCodeAA[c2]:
                
                if Degeneracy[c1][position] == degeneracy and Degeneracy[c2][position] == degeneracy:
                    new_seq1.append( c1[position-1] )
                    new_seq2.append( c2[position-1] )
                    
    return string.join(new_seq1, ""), string.join(new_seq2, "")


##-------------------------------------------------------------------
class SequencePairInfo:
    """the first characters are ACGT."""
    def __init__(self):
        self.mNIdentical = 0
        self.mNAligned = 0
        self.mNDifferent = 0
        self.mNTransitions = 0
        self.mNTransversions = 0
        self.mNUnaligned1 = 0
        self.mNUnaligned2 = 0        
        self.mAlphabet = 0
        self.mMapChar2Pos = {}
        self.mMatrix = []

    def getGCContent(self):
        """return GC content."""
        if not self.mNAligned:
            raise ValueError, "no data for calculating GC content"

        cp = self.mMapChar2Pos['C']
        gp = self.mMapChar2Pos['G']
        # two row and two columns, subtract to avoid double counting
        gc = numpy.sum(self.mMatrix[0:4,cp] \
             + self.mMatrix[0:4,gp] \
             + self.mMatrix[cp,0:4] \
             + self.mMatrix[gp,0:4]) \
             - self.mMatrix[cp,cp] \
             - self.mMatrix[gp,gp] \
             - self.mMatrix[cp,gp] \
             - self.mMatrix[gp,cp] 
             
        return float(gc) / 2.0 / self.mNAligned

    def __str__(self):
        return "\t".join(map(str, (self.mNIdentical, self.mNAligned, self.mNDifferent,
                                   self.mNTransitions, self.mNTransversions,
                                   self.mNUnaligned1, self.mNUnaligned2 )) )

    def getHeader( self ):
        return "\t".join( ("identical", "aligned", "different", "transitions", "transversions",
                             "unaligned1", "unaligned2" ) )

class SequencePairInfoCodons( SequencePairInfo):
    """the first characters are ACGT."""
    def __init__(self):
        SequencePairInfo.__init__(self)
        self.mNNonSynonymous = 0
        self.mNSynonymous = 0

    def __str__(self):

        return "\t".join( (SequencePairInfo.__str__(self),
                           str(self.mNNonSynonymous), 
                           str(self.mNSynonymous )))

    def getHeader( self ):
        return "\t".join( (SequencePairInfo.getHeader(self), "nnonsyn", "nsyn" ))

##-------------------------------------------------------------------------
def AlignedPair2SubstitutionMatrix( seq1, seq2, alphabet ):
    """given a pair of sequences, calculate
    a substitution matrix for the given alphabet.
    """
    if len(seq1) != len(seq2):
        raise "two sequences of unequal length submitted."

    nchars = len(alphabet)
    matrix = numpy.zeros( (nchars,nchars), numpy.int)

    map_char2pos = {}
    for x in alphabet: map_char2pos[x] = len(map_char2pos)
        
    for x in range(len(seq1)):
        try:
            matrix[map_char2pos[seq1[x]],map_char2pos[seq2[x]]] += 1
        except KeyError:
            continue

    return matrix
    
##-------------------------------------------------------------------
def CalculatePairIndices( seq1, seq2, gap_char = "-", with_codons = False ):
    """returns number of idential and transitions/transversions substitutions
    in the alignment.
    
    If with-codons = True, synonymous and nonsynonymous changes will be recorded
    as well. The routine assumes no frame-shifts and will count more than one change
    as non-synonymous.
    """
    alphabet = "ACGT" + gap_char
    
    map_char2pos = {}
    for x in alphabet: map_char2pos[x] = len(map_char2pos)

    ## build coordinates for various substitution subsets
    transitions, transversions = [], []
    for x in ("AG", "GA", "CT", "TC"):
        transitions.append( (map_char2pos[x[0]], map_char2pos[x[1]]) )

    for x in ("AT", "TA", "GT", "TG", "GC", "CG", "AC", "CA" ):
        transversions.append( (map_char2pos[x[0]], map_char2pos[x[1]]) )

    matrix = AlignedPair2SubstitutionMatrix( seq1, seq2, alphabet )
    matrix_acgt = matrix[0:4,0:4]

    if with_codons:
        result = SequencePairInfoCodons()
    else:
        result = SequencePairInfo()

    result.mMatrix = matrix
    result.mMapChar2Pos = map_char2pos
    result.mNAligned = numpy.sum(numpy.ravel(matrix_acgt) )
    result.mNIdentical = numpy.sum( numpy.trace(matrix_acgt) )
    result.mNTransitions = numpy.sum( [ matrix[x] for x in transitions ] )
    result.mNTransversions = numpy.sum( [ matrix[x] for x in transversions ] )
    result.mNDifferent = result.mNAligned - result.mNIdentical
    result.mNUnaligned1 = numpy.sum(numpy.ravel(matrix[0:4,4]))    
    result.mNUnaligned2 = numpy.sum(numpy.ravel(matrix[4,0:4]))    

    if with_codons:
        nsyn, nnon = 0, 0
        pairs = zip( seq1, seq2)
        for x in range(len(pairs)):
            a,b = pairs[x]
            if a != b:

                l = (x // 3) * 3
                c1 = MapCodon2AA(seq1[l:l+3])
                c2 = MapCodon2AA(seq2[l:l+3])
                
                if c1 == GAP_CHAR or c2 == GAP_CHAR: continue
                
                # print x, a, b, l, c1, c2, seq1[l:l+3], seq2[l:l+3], c1 == c2           
                if c1 == c2:
                    nsyn += 1
                else:
                    nnon += 1 

        result.mNSynonymous = nsyn
        result.mNNonSynonymous = nnon

    return result

##------------------------------------------------------------
def makeSubstitutionMatrix( type = "EMBOSS" ):
    """make alignator with DNA substitution matrix.

    EMBOSS style matrix:
    identity = 5
    mismatch = -4
    gop = -16
    gep = -4

    ClustalW style matrix:
    match = 1 mismatch = 0
    gop = -10 gep = -0.1
    """

    matrix = []
    
    if type == "emboss":
        match = 5
        mismatch = -4
        gop = -16.0
        gep = -4.0
    elif type == "blastn":
        match = 1
        mismatch = -3
        gop = -5.0
        gep = -2.0
    elif type == "clustal":
        match = 1
        mismatch = 0
        gop = -10.0
        gep = -0.1
    elif type == "iub":
        match = 1.9
        mismatch = 0
        gop = -10.0
        gep = -0.1
    elif type == "trans":
        match = 1
        mismatch = -5
        gop = -10.0
        gep = -0.1
    elif type == "backtrans":
        # mismatches have to be penalized - there should be none.
        match = 1
        mismatch = -10
        gop = -10.0
        gep = -10.0
    else:
        raise "unknown MATRIX."
    
    # build empty matrix
    for m in range(21): matrix.append( [mismatch] * 21 )
        
    # A <-> A
    matrix[0][0] = match
    # C <-> C
    matrix[1][1] = match
    # G <-> G
    matrix[5][5] = match
    # T <-> T
    matrix[16][16] = match

    if type == "trans":
        # A <-> G
        matrix[0][5] = matrix[5][0] = match
        # C <-> T
        matrix[1][16] = matrix[16][1] = match
    elif type == "backtrans":
        # AGCT <-> W
        for x in (0,1,5,16):
            matrix[x][18] = matrix[18][x] = match
        # AG <-> V
        for x in (0,5):
            matrix[x][17] = matrix[17][x] = match
        # CT <-> Y
        for x in (1,16):
            matrix[x][19] = matrix[19][x] = match
        # N <-> N
        for x in (0,1,5,11,16,17,19):
            matrix[x][11] = matrix[11][x] = match
            
    handle_tmpfile, filename_tmpfile = tempfile.mkstemp()
    for m in matrix:
        os.write( handle_tmpfile, string.join( map(str, m), "\t") + "\n")
    os.close(handle_tmpfile)

    smatrix = alignlib.readSubstitutionMatrixAA( filename_tmpfile )
    os.remove( filename_tmpfile)
    return smatrix, gop, gep

##------------------------------------------------------------
def CalculateRCSUValuesFromCounts( counts, pseudo_counts = 0 ):
    """calculate RCSU values for codons.

    RCSU = relative frequency / uniform frequency
    """
    pass

##------------------------------------------------------------
def CalculateCodonFrequenciesFromCounts( counts, pseudo_counts = 0 ):
    """calculate codon frequencies from codon counts per amino acid.
    pseudo_counts are added if desired.
    """
    map_aa2codons = GetMapAA2Codons()
    weights = {}
    
    for aa, codons in map_aa2codons.items():

        total_counts = reduce(lambda x,y: x+y, ( counts[x] + pseudo_counts for x in codons ))
        if total_counts > 0:
            for x in codons:
                weights[x] = float(counts[x] + pseudo_counts) / total_counts
        else:
            for x in codons:            
                weights[x] = 0.0
            
    for codon in StopCodons:
        weights[codon] = 0.0
        
    return weights

##------------------------------------------------------------
def CalculateCAIWeightsFromCounts( counts, pseudo_counts = 0 ):
    """calculate CAI weights from codon counts.
    pseudo_counts are added if desired.
    """
    map_aa2codons = GetMapAA2Codons()
    weights = {}
    
    for aa, codons in map_aa2codons.items():

        max_counts = max( counts[x] for x in codons ) + pseudo_counts
        
        for x in codons:
            weights[x] = float(counts[x] + pseudo_counts) / float(max_counts)

    for codon in StopCodons:
        weights[codon] = 0.0
        
    return weights

##------------------------------------------------------------
def IsJunk( contig ):
    """returns true, if contigs is likely to be junk.

    This is done by name matching. Junk contigs contain either
    one of the following:

    random, unknown, chrU, chU.
    """

    c = contig.lower()
    negative_list = "random", "unknown", "chru", "chu"

    for x in negative_list:
        if re.search( x, c):
            return True
    
    return False

##------------------------------------------------------------
def CountCodons( sequence ):
    """count the codons in a sequence."""
    if len(sequence) % 3 != 0:
        raise "sequence is not multiple of 3: %i" % len(sequence)
                                                      
    counts = {}
    for c in GeneticCodeAA.keys(): counts[c] = 0
    for codon in [ sequence[x:x+3] for x in range(0, len(sequence), 3)]:
        try:
            counts[codon] += 1
        except KeyError:
            ## skip stop-codons
            pass
        
    return counts

##------------------------------------------------------------
def GetUniformCodonUsage():
    """get list of frequencies for codons expected for uniform codon usage."""

    frequencies = {}

    aas = {}
    
    for codon, aa in GeneticCodeAA.items():
        if aa not in aas: aas[aa] = 0
        aas[aa] += 1

    for codon, aa in GeneticCodeAA.items():    
        frequencies[codon] = 1.0 / aas[aa]
        
    return frequencies

##------------------------------------------------------------
def GetBiasedCodonUsage( bias = 1.0):
    """get list of frequencies for codons according to some bias.

    The first codon for each aa is the most biased, all others are less biased.

    The ratio determines the relative bias between the first and all other
    codons. 0.0 is no bias, 1.0 is complete bias.
    """

    frequencies = {}

    aas = {}
    
    for codon, aa in GeneticCodeAA.items():
        if aa not in aas: aas[aa] = 0
        aas[aa] += 1

    for codon, aa in GeneticCodeAA.items():    
        frequencies[codon] = 1.0 / aas[aa]
        
    return frequencies

##------------------------------------------------------------
def IsNegativeStrand( strand ):
    return str(strand) in ("-", "0", "-1" )

##------------------------------------------------------------
def IsPositiveStrand( strand ):
    return not IsNegativeStrand( strand )

##------------------------------------------------------------
def convertStrand( strand ):
    """convert various strand notations into [+-.].
    """
    s = str(strand) 
    if s in ("-", "0", "-1" ):
        return "-"
    elif s in ("+", "1" ):
        return "+"
    else:
        return "."

##------------------------------------------------------------
def IsJunk( contig ):
    """returns true, if contigs is likely to be junk.

    This is done by name matching. Junk contigs contain either
    one of the following:

    random, unknown, chrU, chU.
    """

    c = contig.lower()
    negative_list = "random", "unknown", "chru", "chu"

    for x in negative_list:
        if re.search( x, c):
            return True
    
    return False

##------------------------------------------------------------
INTRON_TYPES = ( ( "U2-GT/AG", "GT", "AG"),
                 ( "U2-nc-GC/AG", "GC", "AG"),
                 ( "U12-AT/AC", "AT", "AC") )

def GetIntronType( sequence, both_strands = False ):
    """return intron type for an intronic sequence.

    If both_strands == True, both strands are checked.
    """

    if both_strands == False:
        for name, prime5, prime3 in INTRON_TYPES:
            if sequence[:len(prime5)].upper() == prime5 and \
                    sequence[-len(prime3):].upper() == prime3:
                return name, prime5, prime3
        else:
            return "unknown", sequence[:5], sequence[-5:]
    else:
        r = complement( sequence )

        for name, prime5, prime3 in INTRON_TYPES:
            if sequence[:len(prime5)].upper() == prime5 and \
                    sequence[-len(prime3):].upper() == prime3:
                return name, prime5, prime3
            elif r[:len(prime5)].upper() == prime5 and \
                    r[-len(prime3):].upper() == prime3:
                return name, prime3, prime5
        else:
            return "unknown", sequence[:5], sequence[-5:]

##------------------------------------------------------------
def printPrettyAlignment( seq1, *args ):
    """print a pretty alignment."""

    for other in args:
        assert len(seq1) == len(other)

    seqrow, otherrows = [], []
    nother = len(args)
    for x in range(nother):
        otherrows.extend( [ [], [] ] )

    for p in range( len(seq1 )):
        if p > 0 and p % 3 == 0:
            seqrow.append( " ")
            for x in range(2*nother):
                otherrows[x] += " "
            if len(seqrow) > 120:
                print "".join(seqrow)
                for x in range(2*nother):
                    print "".join(otherrows[x])

                print
                seqrow, otherrows = [], []
                nother = len(args)
                for x in range(nother):
                    otherrows.extend( [ [], [] ] )
                    
            
        l1 = len(seq1[p])
        c1 = seq1[p]
        max_len = max( [len(x[p]) for x in args] )
        seqrow.append( c1 + " " * (max_len - 11) )

        for x in range(nother):

            c2 = args[x][p]
            l2 = len(c2)
            if c1 == c2:
                code = "|"
            elif l1 < l2:
                code = "-"
            elif l2 > l2:
                code = "+"
            else:
                code = ":"
            otherrows[x*2].append( code + " " * (max_len - len(code)))

            if l2 == 0: c2,l2 = "-",1
            otherrows[x*2+1].append( c2 + " " * (max_len-l2) )

    print "".join(seqrow)
    for x in range(2*nother):
        print "".join(otherrows[x])





################################################################################
#
#   MRC FGU Computational Genomics Group
#
#   $Id: cgat_script_template.py 2871 2010-03-03 10:20:44Z andreas $
#
#   Copyright (C) 2009 Andreas Heger
#
#   This program is free software; you can redistribute it and/or
#   modify it under the terms of the GNU General Public License
#   as published by the Free Software Foundation; either version 2
#   of the License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#################################################################################
'''
bam2transcriptContribution.py
=============================================

:Author: Nick Ilott
:Release: $Id$
:Date: |today|
:Tags: Python

Purpose
-------

Computes the total number of reads from a BAM alignment file that contribute to
transcripts provided in a gtf file.



Usage
-----

Example::

   python bam2transcriptContribution.py --help

Type::

   python bam2transcriptContribution.py --help

for command line help.

Documentation
-------------

Code
----

'''

import os
import sys
import re
import optparse
import collections
import CGAT.Experiment as E
import CGAT.IOTools as IOTools
import pysam
import CGAT.GTF as GTF
import numpy
import pybedtools
import tempfile
import pyximport
pyximport.install(build_in_temp=False)
import CGAT.Pipeline as P

#########################################################
#########################################################
#########################################################

def main( argv = None ):
    """script main.

    parses command line options in sys.argv, unless *argv* is given.
    """

    if not argv: argv = sys.argv

    # setup command line parser
    parser = E.OptionParser( version = "%prog version: $Id: cgat_script_template.py 2871 2010-03-03 10:20:44Z andreas $", 
                                    usage = globals()["__doc__"] )

    parser.add_option("-b", "--bam-file", dest="bam_file", type="string",
                      help="supply input bam file name"  )
   
    parser.add_option("-g", "--gtf-file", dest="gtf_file", type="string",
                      help="supply input gtf file name"  )
   
    parser.add_option("-o", "--outfile", dest = "outfile", type = "string",
                      help="supply output file name")

    parser.add_option("-G", "--reference-GTF", dest = "reference_gtf", type = "string",
                      help="supply reference gtf for context of reads not contributing to transcripts")


    ## add common options (-h/--help, ...) and parse command line 
    (options, args) = E.Start( parser, argv = argv )

    ######################################################
    ######################################################
    # for all alignments
    ######################################################
    ######################################################

    # open outfile and prepare headers
    outf = open(options.outfile, "w")
    outf.write("\t".join(["total alignments"
                          , "aligments in transcripts"
                          , "percent alignments in transcripts"
                          , "total spliced alignments"
                          , "spliced alignments in transcripts"
                          , "percent spliced alignments in transcripts"]) + "\n")

    # calculate coverage over transcript file - NB split reads contribute twice to the transcript
    # use BedTool object
    pybedbamfile = pybedtools.BedTool(options.bam_file)
    
    # count alignments 
    E.info("counting total number of alignments and spliced alignments")
    total_alignments = 0
    spliced_alignments = 0
    
    for alignment in pybedbamfile:
        cigar = alignment[5]
        if cigar.find("N") != -1: # N signifies split read 
            total_alignments += 1
            spliced_alignments += 1
        else:
            total_alignments += 1

    # merge the gtf file to avoid double counting of exons in different transcripts - converts to a bed file
    gtffile = pybedtools.BedTool(options.gtf_file).merge()

    E.info("computing coverage of aligments in %s over intervals in %s" % (options.bam_file, options.gtf_file))
    cover = pybedbamfile.coverage(gtffile)

    # make sure that the exons aren't being counted twice - shouldn't be because of merge
    E.info("counting reads contributing to transcripts")
    c = 0
    for entry in cover:
        coverage = int(entry[3])
        if coverage > 0:
            c+=coverage
        
    # sum the coverage across exons from all transcripts
    coverage_in_transcripts = c

    ######################################################
    ######################################################
    # for spliced alignments
    ######################################################
    ######################################################

    # count total number of spliced alignments
    # requires that the CIGAR string 'N' is present
    
    # uses pysam to write out a bam file of the spliced reads only
    allreads = pysam.Samfile(options.bam_file)
    spliced_bamname = P.snip(options.bam_file, ".bam") + "_spliced_reads.bam"
    
    # open file for outputting spliced alignments
    splicedreads = pysam.Samfile(spliced_bamname, "wb", template=allreads)
    
    # cigar string in pysam for spliced alignment is (3, int)
    spliced = collections.defaultdict(list)
    for read in allreads:
        for cigar_tag in read.cigar:
            if cigar_tag[0] == 3:
                spliced[read].append(cigar_tag)

    # write out spliced alignments
    for read in spliced.keys():
        splicedreads.write(read)
    splicedreads.close()
    allreads.close()
    
    # index splice reads bam file
    pysam.sort(spliced_bamname, P.snip(spliced_bamname, ".bam"))
    pysam.index(spliced_bamname)

    # read in the spliced reads as a BedTool object
    splicedbam = pybedtools.BedTool(spliced_bamname)

    # perform coverage of spliced reads over intervals - will be twice as many as there should be
    # due to counting both exons overlapping
    spliced_coverage = splicedbam.coverage(gtffile)

    # avoid double counting exons
    E.info("counting spliced reads contributing to transcripts")
    spliced_exons = {}
    c = 0
    for entry in spliced_coverage:
        coverage = int(entry[3])
        if coverage > 0:
            c+=coverage
                
    spliced_coverage_in_transcripts = c

    # NOTE: the counting of spliced alignments is not accurate

    spliced_coverage_in_transcripts =  float(spliced_coverage_in_transcripts)/2
    
    ###########################
    ## write out the results ##
    ###########################

    outf.write(str(int(total_alignments)) + "\t")
    # remove half of the coverage assigned to spliced reads
    coverage_in_transcripts = (coverage_in_transcripts)-(spliced_coverage_in_transcripts)
    outf.write(str(int(coverage_in_transcripts)-int(spliced_coverage_in_transcripts)) + "\t")
    outf.write( str( int((coverage_in_transcripts/total_alignments) *100)) + "\t" )

    # write out spliced counts
    outf.write(str(int(spliced_alignments)) + "\t")
    outf.write(str(int(spliced_coverage_in_transcripts)) + "\t")
    outf.write( str( int( (spliced_coverage_in_transcripts/spliced_alignments) *100)) )

    outf.close()

    ############################
    # contextualise those that 
    # don't fall in transcripts
    ############################

    if options.reference_gtf:
        context_summary = open(P.snip(options.bam_file, ".bam") + ".excluded.context", "w")
        context_summary.write("\t".join(["Feature", "number"]) + "\n")

        # write out the read info as well
        context_file = open(P.snip(options.bam_file, ".bam") + ".excluded", "w")
    
        context_dict = collections.defaultdict(int)
        # intersect bam - write non-overlapping with transcripts - intersect with reference - write out
        context = pybedbamfile.intersect(gtffile, v=True, bed=True).intersect(pybedtools.BedTool(options.reference_gtf), wb=True)
        for entry in context:
            feature = entry[8]
            context_dict[feature] += 1
            context_file.write("\t".join([e for e in entry]) + "\n")

        for feature, value in context_dict.iteritems():
            context_summary.write("\t".join([feature, str(value)]) + "\n")

        context_file.close()
        context_summary.close()


    ## write footer and output benchmark information.
    E.Stop()


if __name__ == "__main__":
    sys.exit( main( sys.argv) )





    




    

